from collections import defaultdict
from itertools import product
from typing import (
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import attr
import numpy as np

from .mag import Mag
from .vec3_int import Vec3Int, Vec3IntLike
from .vec_int import VecInt, VecIntLike

_DEFAULT_BBOX_NAME = "Unnamed Bounding Box"

_T = TypeVar("_T", bound="NDBoundingBox")


def str_tpl(str_list: Iterable[str]) -> Tuple[str, ...]:
    # Fix for mypy bug https://github.com/python/mypy/issues/5313.
    # Solution based on other issue for the same bug: https://github.com/python/mypy/issues/8389.
    return tuple(str_list)


def int_tpl(vec_int_like: VecIntLike) -> VecInt:
    return VecInt(
        vec_int_like, axes=(f"unset_{i}" for i in range(len(list(vec_int_like))))
    )


@attr.frozen
class NDBoundingBox:
    """
    The NDBoundingBox class is a generalized version of the 3-dimensional BoundingBox class. It is designed to represent bounding boxes in any number of dimensions.

    The bounding box is characterized by its top-left corner, the size of the box, the names of the axes for each dimension, and the index (or order) of the axes. Each axis must have a unique index, starting from 1 (index 0 is reserved for channel information).

    The top-left coordinate is inclusive, while the bottom-right coordinate is exclusive.

    Here's a brief example of how to use it:

    ```python

    # Create a 2D bounding box
    bbox_1 = NDBoundingBox(
        top_left=(0, 0),
        size=(100, 100),
        axes=("x", "y"),
        index=(1,2)
    )

    # Create a 4D bounding box
    bbox_2 = NDBoundingBox(
        top_left=(75, 75, 75, 0),
        size=(100, 100, 100, 20),
        axes=("x", "y", "z", "t"),
        index=(2,3,4,1)
    )
    ```
    """

    topleft: VecInt = attr.field(converter=int_tpl)
    size: VecInt = attr.field(converter=int_tpl)
    axes: Tuple[str, ...] = attr.field(converter=str_tpl)
    index: VecInt = attr.field(converter=int_tpl)
    bottomright: VecInt = attr.field(init=False)
    name: Optional[str] = _DEFAULT_BBOX_NAME
    is_visible: bool = True
    color: Optional[Tuple[float, float, float, float]] = None

    def __attrs_post_init__(self) -> None:
        assert (
            len(self.topleft) == len(self.size) == len(self.axes) == len(self.index)
        ), (
            f"The dimensions of topleft, size, axes and index ({len(self.topleft)}, "
            + f"{len(self.size)}, {len(self.axes)} and {len(self.index)}) do not match."
        )
        assert 0 not in self.index, "Index 0 is reserved for channels."

        # Convert the delivered tuples to VecInts
        object.__setattr__(self, "topleft", VecInt(self.topleft, axes=self.axes))
        object.__setattr__(self, "size", VecInt(self.size, axes=self.axes))
        object.__setattr__(self, "index", VecInt(self.index, axes=self.axes))

        if not self._is_sorted():
            self._sort_positions_of_axes()

        if not self.size.is_positive():
            # Flip the size in negative dimensions, so that the topleft is smaller than bottomright.
            # E.g. BoundingBox((10, 10, 10), (-5, 5, 5)) -> BoundingBox((5, 10, 10), (5, 5, 5)).
            negative_size = tuple(min(0, value) for value in self.size)
            new_topleft = tuple(
                val1 + val2 for val1, val2 in zip(self.topleft, negative_size)
            )
            new_size = (abs(value) for value in self.size)
            object.__setattr__(self, "topleft", VecInt(new_topleft, axes=self.axes))
            object.__setattr__(self, "size", VecInt(new_size, axes=self.axes))

        # Compute bottomright to avoid that it's recomputed every time
        # it is needed.
        object.__setattr__(
            self,
            "bottomright",
            self.topleft + self.size,
        )

    def _sort_positions_of_axes(self) -> None:
        # Bring topleft and size in required order
        # defined in axisOrder and index of additionalAxes

        size, topleft, axes, index = zip(
            *sorted(
                zip(self.size, self.topleft, self.axes, self.index), key=lambda x: x[3]
            )
        )
        object.__setattr__(self, "size", VecInt(size, axes=axes))
        object.__setattr__(self, "topleft", VecInt(topleft, axes=axes))
        object.__setattr__(self, "axes", axes)
        object.__setattr__(self, "index", VecInt(index, axes=axes))

    def _is_sorted(self) -> bool:
        return all(self.index[i - 1] < self.index[i] for i in range(1, len(self.index)))

    def with_name(self: _T, name: Optional[str]) -> _T:
        """
        Returns a new instance of `NDBoundingBox` with the specified name.

        Args:
        - name (Optional[str]): The name to assign to the new `NDBoundingBox` instance.

        Returns:
        - NDBoundingBox: A new instance of `NDBoundingBox` with the specified name.
        """
        return attr.evolve(self, name=name)

    def with_topleft(self: _T, new_topleft: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified top left coordinates.

        Args:
        - new_topleft (VecIntLike): The new top left coordinates for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated top left coordinates.
        """
        return attr.evolve(self, topleft=VecInt(new_topleft, axes=self.axes))

    def with_size(self: _T, new_size: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified size.

        Args:
        - new_size (VecIntLike): The new size of the bounding box. Can be a VecInt or any object that can be converted to a VecInt.

        Returns:
        - A new NDBoundingBox object with the specified size.
        """
        return attr.evolve(self, size=VecInt(new_size, axes=self.axes))

    def with_index(self: _T, new_index: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified index.

        Args:
        - new_index (VecIntLike): The new axis order for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated index.
        """
        axes, _ = zip(*sorted(zip(self.axes, new_index), key=lambda x: x[1]))
        return attr.evolve(self, index=VecInt(new_index, axes=axes))

    def with_bottomright(self: _T, new_bottomright: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox with an updated bottomright value.

        Args:
        - new_bottomright (VecIntLike): The new bottom right corner coordinates.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated bottom right corner.
        """
        new_size = VecInt(new_bottomright, axes=self.axes) - self.topleft

        return self.with_size(new_size)

    def with_is_visible(self: _T, is_visible: bool) -> _T:
        """
        Returns a new NDBoundingBox object with the specified visibility.

        Args:
        - is_visible (bool): The visibility value to set.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated visibility value.
        """
        return attr.evolve(self, is_visible=is_visible)

    def with_color(self: _T, color: Optional[Tuple[float, float, float, float]]) -> _T:
        """
        Returns a new instance of NDBoundingBox with the specified color.

        Args:
        - color (Optional[Tuple[float, float, float, float]]): The color to set for the bounding box.
            The color should be specified as a tuple of four floats representing RGBA values.

        Returns:
        - NDBoundingBox: A new instance of NDBoundingBox with the specified color.
        """
        return attr.evolve(self, color=color)

    def with_bounds(
        self: _T, axis: str, new_topleft: Optional[int], new_size: Optional[int]
    ) -> _T:
        """
        Returns a new NDBoundingBox object with updated bounds along the specified axis.

        Args:
        - axis (str): The name of the axis to update.
        - new_topleft (Optional[int]): The new value for the top-left coordinate along the specified axis.
        - new_size (Optional[int]): The new size along the specified axis.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with updated bounds.

        Raises:
        - ValueError: If the given axis name does not exist.

        """
        try:
            index = self.axes.index(axis)
        except ValueError as err:
            raise ValueError("The given axis name does not exist.") from err

        _new_topleft = (
            self.topleft.with_replaced(index, new_topleft)
            if new_topleft is not None
            else self.topleft
        )
        _new_size = (
            self.size.with_replaced(index, new_size)
            if new_size is not None
            else self.size
        )

        return attr.evolve(self, topleft=_new_topleft, size=_new_size)

    def get_bounds(self, axis: str) -> Tuple[int, int]:
        """
        Returns the bounds of the given axis.

        Args:
        - axis (str): The name of the axis to get the bounds for.

        Returns:
        - Tuple[int, int]: A tuple containing the top-left and bottom-right coordinates along the specified axis.
        """
        try:
            index = self.axes.index(axis)
        except ValueError as err:
            raise ValueError("The given axis name does not exist.") from err

        return (self.topleft[index], self.topleft[index] + self.size[index])

    @classmethod
    def group_boxes_with_aligned_mag(
        cls, bounding_boxes: Iterable["NDBoundingBox"], aligning_mag: Mag
    ) -> Dict["NDBoundingBox", List["NDBoundingBox"]]:
        """
        Groups the given BoundingBox instances by aligning each
        bbox to the given mag and using that as the key.
        For example, bounding boxes of size 256**3 could be grouped
        into the corresponding 1024**3 chunks to which they belong
        by using aligning_mag = Mag(1024).
        """

        chunks_with_bboxes = defaultdict(list)
        for bbox in bounding_boxes:
            chunk_key = bbox.align_with_mag(aligning_mag, ceil=True)
            chunks_with_bboxes[chunk_key].append(bbox)

        return chunks_with_bboxes

    @classmethod
    def from_wkw_dict(cls, bbox: Dict) -> "NDBoundingBox":
        """
        Create an instance of NDBoundingBox from a dictionary representation.

        Args:
        - bbox (Dict): The dictionary representation of the bounding box.

        Returns:
        - NDBoundingBox: An instance of NDBoundingBox.

        Raises:
        - AssertionError: If additionalAxes are present but axisOrder is not provided.
        """

        topleft: Tuple[int, ...] = bbox["topLeft"]
        size: Tuple[int, ...] = (bbox["width"], bbox["height"], bbox["depth"])
        axes: Tuple[str, ...] = ("x", "y", "z")
        index: Tuple[int, ...] = (1, 2, 3)

        if "axisOrder" in bbox:
            axes = tuple(bbox["axisOrder"].keys())
            index = tuple(bbox["axisOrder"][axis] for axis in axes)

            if "additionalAxes" in bbox:
                assert (
                    "axisOrder" in bbox
                ), "If there are additionalAxes an axisOrder needs to be provided."
                for axis in bbox["additionalAxes"]:
                    topleft += (axis["bounds"][0],)
                    size += (axis["bounds"][1] - axis["bounds"][0],)
                    axes += (axis["name"],)
                    index += (axis["index"],)

        return cls(
            topleft=VecInt(topleft, axes=axes),
            size=VecInt(size, axes=axes),
            axes=axes,
            index=VecInt(index, axes=axes),
        )

    def to_wkw_dict(self) -> dict:
        """
        Converts the bounding box object to a json dictionary.

        Returns:
        - dict: A json dictionary representing the bounding box.
        """
        topleft = [None, None, None]
        width, height, depth = None, None, None
        additional_axes = []
        for i, axis in enumerate(self.axes):
            if axis == "x":
                topleft[0] = self.topleft[i]
                width = self.size[i]
            elif axis == "y":
                topleft[1] = self.topleft[i]
                height = self.size[i]
            elif axis == "z":
                topleft[2] = self.topleft[i]
                depth = self.size[i]
            else:
                additional_axes.append(
                    {
                        "name": axis,
                        "bounds": [self.topleft[i], self.bottomright[i]],
                        "index": self.index[i],
                    }
                )
        if additional_axes:
            return {
                "topLeft": topleft,
                "width": width,
                "height": height,
                "depth": depth,
                "additionalAxes": additional_axes,
            }
        return {
            "topLeft": topleft,
            "width": width,
            "height": height,
            "depth": depth,
        }

    def to_config_dict(self) -> dict:
        """
        Returns a dictionary representation of the bounding box.

        Returns:
        - dict: A dictionary representation of the bounding box.
        """
        return {
            "topleft": self.topleft.to_list(),
            "size": self.size.to_list(),
            "axes": self.axes,
        }

    def to_checkpoint_name(self) -> str:
        """
        Returns a string representation of the bounding box that can be used as a checkpoint name.

        Returns:
        - str: A string representation of the bounding box.
        """
        return f"{'_'.join(str(element) for element in self.topleft)}_{'_'.join(str(element) for element in self.size)}"

    def __repr__(self) -> str:
        return f"NDBoundingBox(topleft={self.topleft.to_tuple()}, size={self.size.to_tuple()}, axes={self.axes})"

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NDBoundingBox):
            self._check_compatibility(other)
            return self.topleft == other.topleft and self.size == other.size

        raise NotImplementedError()

    def __len__(self) -> int:
        return len(self.axes)

    def get_shape(self, axis_name: str) -> int:
        """
        Returns the size of the bounding box along the specified axis.

        Args:
        - axis_name (str): The name of the axis to get the size for.

        Returns:
        - int: The size of the bounding box along the specified axis.
        """
        try:
            index = self.axes.index(axis_name)
            return self.size[index]
        except ValueError as err:
            raise ValueError(
                f"Axis {axis_name} doesn't exist in NDBoundingBox."
            ) from err

    def _get_attr_xyz(self, attr_name: str) -> Vec3Int:
        axes = ("x", "y", "z")
        attr_3d = []

        for axis in axes:
            index = self.axes.index(axis)
            attr_3d.append(getattr(self, attr_name)[index])

        return Vec3Int(attr_3d)

    def _get_attr_with_replaced_xyz(self, attr_name: str, xyz: Vec3IntLike) -> VecInt:
        value = Vec3Int(xyz)
        axes = ("x", "y", "z")
        modified_attr = getattr(self, attr_name).to_list()

        for i, axis in enumerate(axes):
            index = self.axes.index(axis)
            modified_attr[index] = value[i]

        return VecInt(modified_attr, axes=self.axes)

    @property
    def topleft_xyz(self) -> Vec3Int:
        """The topleft corner of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("topleft")

    @property
    def size_xyz(self) -> Vec3Int:
        """The size of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("size")

    @property
    def bottomright_xyz(self) -> Vec3Int:
        """The bottomright corner of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("bottomright")

    @property
    def index_xyz(self) -> Vec3Int:
        """The index of x, y and z axis within the bounding box."""

        return self._get_attr_xyz("index")

    def with_topleft_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z coordinates of the topleft corner.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z coordinates for the topleft corner.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z coordinates of the topleft corner.
        """
        new_topleft = self._get_attr_with_replaced_xyz("topleft", new_xyz)

        return self.with_topleft(new_topleft)

    def with_size_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z size.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z size for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z size.
        """
        new_size = self._get_attr_with_replaced_xyz("size", new_xyz)

        return self.with_size(new_size)

    def with_bottomright_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z coordinates of the bottomright corner.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z coordinates for the bottomright corner.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z coordinates of the bottomright corner.
        """
        new_bottomright = self._get_attr_with_replaced_xyz("bottomright", new_xyz)

        return self.with_bottomright(new_bottomright)

    def with_index_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z index.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z index for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z index.
        """
        new_index = self._get_attr_with_replaced_xyz("index", new_xyz)

        return self.with_index(new_index)

    def _check_compatibility(self, other: "NDBoundingBox") -> None:
        """Checks if two bounding boxes are comparable. To be comparable they need the same number of axes, with same names and same order."""

        if self.axes != other.axes:
            raise ValueError(
                f"Operation with two bboxes is only possible if they have the same axes and axes order. {self.axes} != {other.axes}"
            )

    def padded_with_margins(
        self, margins_left: VecIntLike, margins_right: Optional[VecIntLike] = None
    ) -> "NDBoundingBox":
        raise NotImplementedError()

    def intersected_with(self: _T, other: _T, dont_assert: bool = False) -> _T:
        """
        Returns the intersection of two bounding boxes.

        If dont_assert is set to False, this method may return empty bounding boxes (size == (0, 0, 0))

        Args:
        - other (NDBoundingBox): The other bounding box to intersect with.
        - dont_assert (bool): If True, the method may return empty bounding boxes.

        Returns:
        - NDBoundingBox: The intersection of the two bounding boxes.
        """

        self._check_compatibility(other)
        topleft = self.topleft.pairmax(other.topleft)
        bottomright = self.bottomright.pairmin(other.bottomright)
        size = (bottomright - topleft).pairmax(VecInt.zeros(self.axes))

        intersection = attr.evolve(self, topleft=topleft, size=size)

        if not dont_assert:
            assert (
                not intersection.is_empty()
            ), f"No intersection between bounding boxes {self} and {other}."

        return intersection

    def extended_by(self: _T, other: _T) -> _T:
        """
        Returns the smallest bounding box that contains both bounding boxes.

        Args:
        - other (NDBoundingBox): The other bounding box to extend with.

        Returns:
        - NDBoundingBox: The smallest bounding box that contains both bounding boxes.
        """
        self._check_compatibility(other)
        if self.is_empty():
            return other
        if other.is_empty():
            return self

        topleft = self.topleft.pairmin(other.topleft)
        bottomright = self.bottomright.pairmax(other.bottomright)
        size = bottomright - topleft

        return attr.evolve(self, topleft=topleft, size=size)

    def is_empty(self) -> bool:
        """
        Boolean check whether the boundung box is empty.

        Returns:
        - bool: True if the bounding box is empty, False otherwise.
        """
        return not self.size.is_positive(strictly_positive=True)

    def in_mag(self: _T, mag: Mag) -> _T:
        """
        Returns the bounding box in the given mag.

        Args:
        - mag (Mag): The magnification to convert the bounding box to.

        Returns:
        - NDBoundingBox: The bounding box in the given magnification.
        """
        mag_vec = mag.to_vec3_int()

        assert (
            self.topleft_xyz % mag_vec == Vec3Int.zeros()
        ), f"topleft {self.topleft} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."
        assert (
            self.bottomright_xyz % mag_vec == Vec3Int.zeros()
        ), f"bottomright {self.bottomright} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."

        return self.with_topleft_xyz(self.topleft_xyz // mag_vec).with_size_xyz(
            self.size_xyz // mag_vec
        )

    def from_mag_to_mag1(self: _T, from_mag: Mag) -> _T:
        """
        Returns the bounging box in the finest magnification (Mag(1)).

        Args:
        - from_mag (Mag): The current magnification of the bounding box.

        Returns:
        - NDBoundingBox: The bounding box in the given magnification.
        """
        mag_vec = from_mag.to_vec3_int()

        return self.with_topleft_xyz(self.topleft_xyz * mag_vec).with_size_xyz(
            self.size_xyz * mag_vec
        )

    def _align_with_mag_slow(self: _T, mag: Mag, ceil: bool = False) -> _T:
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrinked when necessary.
        """
        np_mag = mag.to_np()

        align = (  # noqa E731
            lambda point, round_fn: round_fn(point.to_np() / np_mag).astype(int)
            * np_mag
        )

        if ceil:
            topleft = align(self.topleft, np.floor)
            bottomright = align(self.bottomright, np.ceil)
        else:
            topleft = align(self.topleft, np.ceil)
            bottomright = align(self.bottomright, np.floor)
        return attr.evolve(self, topleft=topleft, size=bottomright - topleft)

    def align_with_mag(self: _T, mag: Union[Mag, Vec3Int], ceil: bool = False) -> _T:
        """
        Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        Args:
        - mag (Union[Mag, Vec3Int]): The magnification to align the bounding box to.
        - ceil (bool): If True, the bounding box is enlarged when necessary. If False, it's shrinked when necessary.

        Returns:
        - NDBoundingBox: The aligned bounding box.
        """
        # This does the same as _align_with_mag_slow, which is more readable.
        # Same behavior is asserted in test_align_with_mag_against_numpy_implementation
        mag_vec = mag.to_vec3_int() if isinstance(mag, Mag) else mag
        topleft = self.topleft_xyz
        bottomright = self.bottomright_xyz
        roundup = topleft if ceil else bottomright
        rounddown = bottomright if ceil else topleft
        margin_to_roundup = roundup % mag_vec
        aligned_roundup = roundup - margin_to_roundup
        margin_to_rounddown = (mag_vec - (rounddown % mag_vec)) % mag_vec
        aligned_rounddown = rounddown + margin_to_rounddown
        if ceil:
            return self.with_topleft_xyz(aligned_roundup).with_size_xyz(
                aligned_rounddown - aligned_roundup
            )
        else:
            return self.with_topleft_xyz(aligned_rounddown).with_size_xyz(
                aligned_roundup - aligned_rounddown
            )

    def contains(self, coord: VecIntLike) -> bool:
        """
        Check whether a point is inside of the bounding box.
        Note that the point may have float coordinates in the ndarray case

        Args:
        - coord (VecIntLike): The coordinates to check.

        Returns:
        - bool: True if the point is inside of the bounding box, False otherwise.
        """

        if isinstance(coord, np.ndarray):
            assert (
                coord.shape == (len(self.size),)
            ), f"Numpy array BoundingBox.contains must have shape ({len(self.size)},), got {coord.shape}."
            return cast(
                bool,
                np.all(coord >= self.topleft) and np.all(coord < self.bottomright),
            )
        else:
            # In earlier versions, we simply converted to ndarray to have
            # a unified calculation here, but this turned out to be a performance bottleneck.
            # Therefore, the contains-check is performed on the tuple here.
            coord = VecInt(coord, axes=self.axes)
            return all(
                self.topleft[i] <= coord[i] < self.bottomright[i]
                for i in range(len(self.axes))
            )

    def contains_bbox(self: _T, inner_bbox: _T) -> bool:
        """
        Check whether a bounding box is completely inside of the bounding box.

        Args:
        - inner_bbox (NDBoundingBox): The bounding box to check.

        Returns:
        - bool: True if the bounding box is completely inside of the bounding box, False otherwise.
        """
        self._check_compatibility(inner_bbox)
        return inner_bbox.intersected_with(self, dont_assert=True) == inner_bbox

    def chunk(
        self: _T,
        chunk_shape: VecIntLike,
        chunk_border_alignments: Optional[VecIntLike] = None,
    ) -> Generator[_T, None, None]:
        """
        Decompose the bounding box into smaller chunks of size `chunk_shape`.

        Chunks at the border of the bounding box might be smaller than chunk_shape.
        If `chunk_border_alignment` is set, all border coordinates
        *between two chunks* will be divisible by that value.

        Args:
        - chunk_shape (VecIntLike): The size of the chunks to generate.
        - chunk_border_alignments (Optional[VecIntLike]): The alignment of the chunk borders.


        Yields:
        - Generator[NDBoundingBox]: A generator of the chunks.
        """

        start = self.topleft.to_np()
        try:
            # If a 3D chunk_shape is given it is assumed that iteration over xyz is
            # intended. Therefore NDBoundingBoxes are generated that have a shape of
            # x: chunk_shape.x, y: chunk_shape.y, z: chunk_shape.z and 1 for all other
            # axes.
            chunk_shape = Vec3Int(chunk_shape)

            chunk_shape = (
                self.with_size(VecInt.ones(self.axes))
                .with_size_xyz(chunk_shape)
                .size.to_np()
            )
        except AssertionError:
            chunk_shape = VecInt(chunk_shape, axes=self.axes).to_np()

        start_adjust = VecInt.zeros(self.axes).to_np()
        if chunk_border_alignments is not None:
            try:
                chunk_border_alignments = Vec3Int(chunk_border_alignments)

                chunk_border_alignments = (
                    self.with_size(VecInt.ones(self.axes))
                    .with_size_xyz(chunk_border_alignments)
                    .size.to_np()
                )
            except AssertionError:
                chunk_border_alignments = VecInt(
                    chunk_border_alignments, axes=self.axes
                ).to_np()

            assert np.all(
                chunk_shape % chunk_border_alignments == 0
            ), f"{chunk_shape} not divisible by {chunk_border_alignments}"

            # Move the start to be aligned correctly. This doesn't actually change
            # the start of the first chunk, because we'll intersect with `self`,
            # but it'll lead to all chunk borders being aligned correctly.
            start_adjust = start % chunk_border_alignments
        for coordinates in product(
            *[
                range(
                    start[i] - start_adjust[i], start[i] + self.size[i], chunk_shape[i]
                )
                for i in range(len(self.axes))
            ]
        ):
            yield self.intersected_with(
                self.__class__(
                    topleft=VecInt(coordinates, axes=self.axes),
                    size=VecInt(chunk_shape, axes=self.axes),
                    axes=self.axes,
                    index=self.index,
                )
            )

    def volume(self) -> int:
        """
        Returns the volume of the bounding box.
        """
        return self.size.prod()

    def slice_array(self, array: np.ndarray) -> np.ndarray:
        """
        Returns a slice of the given array that corresponds to the bounding box.
        """
        return array[self.to_slices()]

    def to_slices(self) -> Tuple[slice, ...]:
        """
        Returns a tuple of slices that corresponds to the bounding box.
        """
        return tuple(
            slice(topleft, topleft + size)
            for topleft, size in zip(self.topleft, self.size)
        )

    def offset(self: _T, vector: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified offset.

        Args:
        - vector (VecIntLike): The offset to apply to the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the specified offset.
        """
        try:
            return self.with_topleft_xyz(self.topleft_xyz + Vec3Int(vector))
        except AssertionError:
            return self.with_topleft(self.topleft + VecInt(vector, axes=self.axes))



# ruff: noqa: F401 imported but unused
from .annotation import (
    Annotation,
    AnnotationState,
    AnnotationType,
    SegmentInformation,
    open_annotation,
)
from .annotation_info import AnnotationInfo



from typing import Any, Callable, Dict, Mapping, Type, TypeVar

import cattrs
from attrs import fields as attr_fields
from attrs import has as is_attr_class

from ...utils import snake_to_camel_case

T = TypeVar("T")

custom_converter = cattrs.Converter()

# Structuring and destructuring used for the attrs classes in apiclient.models.
# The server expects and sends camelCase fields, we want snake_case here
# However, the case conversion should happen only for the attrs classes,
# and not for dicts that may contain user data (e.g. user experiences)


def attr_to_camel_case_structure(cl: Type[T]) -> Callable[[Mapping[str, Any], Any], T]:
    return cattrs.gen.make_dict_structure_fn(
        cl,
        custom_converter,
        **{
            a.name: cattrs.gen.override(rename=snake_to_camel_case(a.name))
            # https://github.com/python/mypy/issues/16254
            for a in attr_fields(cl)  # type: ignore[arg-type,misc]
        },
    )


def attr_to_camel_case_unstructure(cl: Type[T]) -> Callable[[T], Dict[str, Any]]:
    return cattrs.gen.make_dict_unstructure_fn(
        cl,
        custom_converter,
        **{
            a.name: cattrs.gen.override(rename=snake_to_camel_case(a.name))
            # https://github.com/python/mypy/issues/16254
            for a in attr_fields(cl)  # type: ignore[arg-type,misc]
        },
    )


custom_converter.register_structure_hook_factory(
    is_attr_class, attr_to_camel_case_structure
)
custom_converter.register_unstructure_hook_factory(
    is_attr_class, attr_to_camel_case_unstructure
)



from typing import Dict, List, Tuple

import attr

from ..client.api_client.models import ApiLoggedTimeGroupedByMonth, ApiUser
from ..client.context import _get_api_client


@attr.frozen
class User:
    """Represents a user of a webknossos instance.
    You can get users via `get_current_user` and `get_all_managed_users`."""

    user_id: str
    email: str
    organization_id: str
    first_name: str
    last_name: str
    created: int
    last_activity: int
    teams: Tuple["Team", ...]
    experiences: Dict[str, int]
    is_active: bool
    is_admin: bool
    is_dataset_manager: bool

    def get_logged_times(self) -> List["LoggedTime"]:
        """Get the logged times of this user.
        Returns a list of `LoggedTime` objects where one represents one month."""
        client = _get_api_client(enforce_auth=True)
        api_logged_times: ApiLoggedTimeGroupedByMonth = client.user_logged_time(
            self.user_id
        )
        return [
            LoggedTime(
                duration_in_seconds=i.duration_in_seconds,
                month=i.payment_interval.month,
                year=i.payment_interval.year,
            )
            for i in api_logged_times.logged_time
        ]

    @classmethod
    def _from_api_user(cls, api_user: ApiUser) -> "User":
        return cls(
            user_id=api_user.id,
            email=api_user.email,
            organization_id=api_user.organization,
            first_name=api_user.first_name,
            last_name=api_user.last_name,
            created=api_user.created,
            last_activity=api_user.last_activity,
            teams=tuple(
                Team(id=team.id, name=team.name, organization_id=api_user.organization)
                for team in api_user.teams
            ),
            experiences=api_user.experiences,
            is_active=api_user.is_active,
            is_admin=api_user.is_admin,
            is_dataset_manager=api_user.is_dataset_manager,
        )

    @classmethod
    def get_by_id(cls, id: str) -> "User":  # noqa: A002 Argument `id` is shadowing a Python builtin
        """Returns the user specified by the passed id if your token authorizes you to see them."""
        client = _get_api_client(enforce_auth=True)
        api_user = client.user_by_id(id)
        print(api_user)
        return cls._from_api_user(api_user)

    @classmethod
    def get_current_user(cls) -> "User":
        """Returns the current user from the authentication context."""
        client = _get_api_client(enforce_auth=True)
        api_user = client.user_current()
        return cls._from_api_user(api_user)

    @classmethod
    def get_all_managed_users(cls) -> List["User"]:
        """Returns all users of whom the current user is admin or team-manager."""
        client = _get_api_client(enforce_auth=True)
        api_users = client.user_list()
        return [cls._from_api_user(i) for i in api_users]


@attr.frozen
class Team:
    id: str
    name: str
    organization_id: str

    @classmethod
    def get_by_name(cls, name: str) -> "Team":
        """Returns the Team specified by the passed name if your token authorizes you to see it."""
        client = _get_api_client(enforce_auth=True)
        api_teams = client.team_list()
        for api_team in api_teams:
            if api_team.name == name:
                return cls(api_team.id, api_team.name, api_team.organization)
        raise KeyError(f"Could not find team {name}.")


@attr.frozen
class LoggedTime:
    duration_in_seconds: int
    year: int
    month: int



from contextlib import closing, contextmanager
from os import PathLike
from pathlib import Path
from typing import Iterator, Set

import numpy as np

from .vendor.dm3 import DM3  # type: ignore[attr-defined]
from .vendor.dm3 import dT_str as DM3_DTYPE_MAPPING  # type: ignore[attr-defined]
from .vendor.dm4 import DM4File  # type: ignore[attr-defined]

try:
    from pims import FramesSequenceND
except ImportError as e:
    raise RuntimeError(
        "Cannot import pims, please install it e.g. using 'webknossos[all]'"
    ) from e


class PimsDm3Reader(FramesSequenceND):
    @classmethod
    def class_exts(cls) -> Set[str]:
        return {"dm3"}

    # class_priority is used in pims to pick the reader with the highest priority.
    # Default is 10, and bioformats priority is 2.
    # See http://soft-matter.github.io/pims/v0.6.1/custom_readers.html#plugging-into-pims-s-open-function
    class_priority = 20

    def __init__(self, path: PathLike) -> None:
        self.path = Path(path)
        super().__init__()
        dm3_file = DM3(self.path)
        self._init_axis("x", dm3_file.width)
        self._init_axis("y", dm3_file.height)
        if dm3_file.depth > 1:
            self._init_axis("z", dm3_file.depth)
            self._register_get_frame(self._get_frame, "zyx")
        else:
            self._register_get_frame(self._get_frame, "yx")

    @property  # potential @cached_property for py3.8+
    def pixel_type(self) -> np.dtype:
        dm3_file = DM3(self.path)
        return np.dtype(DM3_DTYPE_MAPPING[dm3_file._data_type])

    def _get_frame(self, **ind: int) -> np.ndarray:
        del ind
        return DM3(self.path).imagedata


class PimsDm4Reader(FramesSequenceND):
    @classmethod
    def class_exts(cls) -> Set[str]:
        return {"dm4"}

    # class_priority is used in pims to pick the reader with the highest priority.
    # Default is 10, and bioformats priority is 2.
    # See http://soft-matter.github.io/pims/v0.6.1/custom_readers.html#plugging-into-pims-s-open-function
    class_priority = 20

    def __init__(self, path: PathLike) -> None:
        self.path = Path(path)
        super().__init__()
        with self.dm4_file() as dm4_file:
            tags = dm4_file.read_directory()
            image_data_tag = (
                tags.named_subdirs["ImageList"]
                .unnamed_subdirs[1]
                .named_subdirs["ImageData"]
            )
            self._image_tag = image_data_tag.named_tags["Data"]
            self._shape = [
                dm4_file.read_tag_data(i)
                for i in image_data_tag.named_subdirs["Dimensions"].unnamed_tags
            ]

            if len(self._shape) not in [2, 3]:
                raise ValueError(
                    f"DM4 file {self.path} has incompatible number of dimensions, got shape {self._shape}."
                )
            self._init_axis("x", self._shape[0])
            self._init_axis("y", self._shape[1])
            if len(self._shape) == 2:
                self._register_get_frame(self._get_frame, "yx")
            else:
                self._init_axis("z", self._shape[2])
                self._register_get_frame(self._get_frame, "zyx")

    @contextmanager
    def dm4_file(self) -> Iterator[DM4File]:
        with closing(DM4File.open(self.path)) as dm4_file:
            yield dm4_file

    @property  # potential @cached_property for py3.8+
    def pixel_type(self) -> np.dtype:
        with self.dm4_file() as dm4_file:
            return np.dtype(dm4_file.read_tag_data_type(self._image_tag))

    def _get_frame(self, **ind: int) -> np.ndarray:
        del ind
        with self.dm4_file() as dm4_file:
            a = np.asarray(dm4_file.read_tag_data(self._image_tag))
        return a.reshape(self._shape[::-1])



import itertools
from os import PathLike
from pathlib import Path
from typing import Iterator, Optional, Tuple, Union

import attr

from ..utils import warn_deprecated
from .group import Group

Vector3 = Tuple[float, float, float]


@attr.define()
class Skeleton(Group):
    """
    Representation of the [skeleton](/webknossos/skeleton_annotation.html) of an `Annotation`.
    It contains metadata to identify the related dataset and is the root-group of sub-groups and trees.
    See the parent class `Group` for methods about group and tree handling.
    To upload a skeleton to webknossos, please create an `Annotation()` with it.

    A small usage example:

    ```python
    annotation = Annotation(
        name="my_annotation", dataset_name="my_dataset", voxel_size=(11, 11, 24)
    )
    group = annotation.skeleton.add_group("a group")
    tree = group.add_tree("a tree")
    node_1 = tree.add_node(position=(0, 0, 0), comment="node 1")
    node_2 = tree.add_node(position=(100, 100, 100), comment="node 2")

    tree.add_edge(node_1, node_2)
    ```

    Also see [this example](/webknossos-py/examples/skeleton_synapse_candidates.html) for a more
    complex interaction.
    """

    voxel_size: Vector3
    dataset_name: str
    organization_id: Optional[str] = None
    description: Optional[str] = None
    # from Group parent to support mypy:
    _enforced_id: Optional[int] = attr.field(default=None, eq=False, repr=False)

    name: str = attr.field(default="Root", init=False, eq=False, repr=False)
    """Should not be used with `Skeleton`, this attribute is only useful for sub-groups. Set to `Root`."""

    # initialized in post_init:
    _id: int = attr.field(init=False, repr=False)
    _element_id_generator: Iterator[int] = attr.field(init=False, eq=False, repr=False)
    _skeleton: "Skeleton" = attr.field(init=False, eq=False, repr=False)

    @classmethod
    def _set_init_docstring(cls) -> None:
        Skeleton.__init__.__doc__ = """
        To initialize a skeleton, setting the following parameters is required (or recommended):
        - voxel_size
        - dataset_name
        - organization_id
        - description
        """

    def __attrs_post_init__(self) -> None:
        self._element_id_generator = itertools.count()
        self._skeleton = self
        super().__attrs_post_init__()  # sets self._id

    @property
    def scale(self) -> Tuple[float, float, float]:
        """Deprecated, please use `voxel_size`."""
        warn_deprecated("scale", "voxel_size")
        return self.voxel_size

    @scale.setter
    def scale(self, scale: Tuple[float, float, float]) -> None:
        """Deprecated, please use `voxel_size`."""
        warn_deprecated("scale", "voxel_size")
        self.voxel_size = scale

    @staticmethod
    def load(file_path: Union[PathLike, str]) -> "Skeleton":
        """Loads a `.nml` file or a `.zip` file containing an NML (and possibly also volume
        layers). Returns the `Skeleton` object. Also see `Annotation.load` if you want to
        have the annotation which wraps the skeleton."""

        from ..annotation import Annotation

        return Annotation.load(file_path).skeleton

    def save(self, out_path: Union[str, PathLike]) -> None:
        """
        Stores the skeleton as a zip or nml at the given path.
        """

        from ..annotation import Annotation

        out_path = Path(out_path)
        assert out_path.suffix in [
            ".nml",
            ".zip",
        ], f"The suffix if the file must be .nml or .zip, not {out_path.suffix}"
        annotation = Annotation(name=out_path.stem, skeleton=self, time=None)
        annotation.save(out_path)

    @staticmethod
    def from_path(file_path: Union[PathLike, str]) -> "Skeleton":
        """Deprecated. Use Skeleton.load instead."""
        warn_deprecated("Skeleton.from_path", "Skeleton.load")
        return Skeleton.load(file_path)

    def write(self, out_path: PathLike) -> None:
        """Deprecated. Use Skeleton.save instead."""
        warn_deprecated("Skeleton.write", "skeleton.save")
        self.save(out_path)

    def __hash__(self) -> int:
        return id(self)


Skeleton._set_init_docstring()



import colorsys
import itertools
import warnings
from typing import TYPE_CHECKING, List, Tuple

import networkx as nx
import numpy as np

import webknossos._nml as wknml

from ..geometry import Vec3Int

if TYPE_CHECKING:
    from ..annotation import Annotation
    from ..skeleton import Group, Skeleton, Tree


def nml_to_skeleton(nml: wknml.Nml) -> "Skeleton":
    from ..skeleton import Skeleton

    skeleton = Skeleton(
        dataset_name=nml.parameters.name,
        voxel_size=nml.parameters.scale,
        organization_id=nml.parameters.organization,
        description=nml.parameters.description,
    )

    groups_by_id = {}

    def visit_groups(nml_groups: List[wknml.Group], current_group: "Group") -> None:
        for nml_group in nml_groups:
            sub_group = current_group.add_group(
                name=nml_group.name, _enforced_id=nml_group.id
            )
            groups_by_id[sub_group.id] = sub_group
            visit_groups(nml_group.children, sub_group)

    visit_groups(nml.groups, skeleton)
    for nml_tree in nml.trees:
        if nml_tree.groupId is None:
            new_tree = skeleton.add_tree(
                nml_tree.name, _enforced_id=nml_tree.id, color=nml_tree.color
            )
        else:
            new_tree = groups_by_id[nml_tree.groupId].add_tree(
                nml_tree.name, _enforced_id=nml_tree.id, color=nml_tree.color
            )
        _nml_tree_to_wk_tree(new_tree, nml_tree)

    for comment in nml.comments:
        skeleton.get_node_by_id(comment.node).comment = comment.content

    for branchpoint in nml.branchpoints:
        node = skeleton.get_node_by_id(branchpoint.id)
        node.is_branchpoint = True
        if branchpoint.time != 0:
            node.branchpoint_time = branchpoint.time

    max_id = max(skeleton.get_max_tree_id(), skeleton.get_max_node_id())
    skeleton._element_id_generator = itertools.count(max_id + 1)

    if nml.parameters.offset is not None and not all(
        i == 0 for i in nml.parameters.offset
    ):
        warnings.warn(
            f"[INFO] The offset parameter is not supported. The offset of {nml.parameters.offset} is ignored and will be omitted."
        )

    return skeleton


def _nml_tree_to_wk_tree(
    new_tree: "Tree",
    nml_tree: wknml.Tree,
) -> None:
    optional_attribute_list = [
        "rotation",
        "inVp",
        "inMag",
        "bitDepth",
        "interpolation",
        "time",
    ]

    for nml_node in nml_tree.nodes:
        node_id = nml_node.id
        current_node = new_tree.add_node(
            position=Vec3Int.from_vec3_float(nml_node.position),
            _enforced_id=node_id,
            radius=nml_node.radius,
        )

        for optional_attribute in optional_attribute_list:
            if getattr(nml_node, optional_attribute) is not None:
                setattr(
                    current_node,
                    optional_attribute,
                    getattr(nml_node, optional_attribute),
                )

    for edge in nml_tree.edges:
        new_tree.add_edge(edge.source, edge.target)


def _random_color_rgba() -> Tuple[float, float, float, float]:
    """
    A utility to generate a new random RGBA color.
    """
    # https://stackoverflow.com/a/43437435/783758

    h, s, l = (  # noqa: E741 Ambiguous variable name
        np.random.random(),
        0.5 + np.random.random() / 2.0,
        0.4 + np.random.random() / 5.0,
    )
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    return (r, g, b, 1.0)


def annotation_to_nml(
    annotation: "Annotation",
) -> wknml.Nml:
    nmlParameters = wknml.Parameters(
        name=annotation.dataset_name,
        scale=annotation.voxel_size,
        description=annotation.description,
        organization=annotation.organization_id,
        time=annotation.time,
        editPosition=annotation.edit_position,
        editRotation=annotation.edit_rotation,
        zoomLevel=annotation.zoom_level,
        taskBoundingBox=annotation.task_bounding_box,
        userBoundingBoxes=annotation.user_bounding_boxes,
    )

    comments = [
        wknml.Comment(node.id, node.comment)
        for tree in annotation.skeleton.flattened_trees()
        for node in tree.nodes
        if node.comment is not None
    ]

    branchpoints = [
        wknml.Branchpoint(node.id, node.time)
        for tree in annotation.skeleton.flattened_trees()
        for node in tree.nodes
        if node.is_branchpoint
    ]

    trees = []

    for tree in sorted(annotation.skeleton.flattened_trees(), key=lambda g: g.id):
        nodes, edges = _extract_nodes_and_edges_from_tree(tree)
        color = tree.color or _random_color_rgba()
        name = tree.name or f"tree{tree.id}"

        trees.append(
            wknml.Tree(
                nodes=nodes,
                edges=edges,
                id=tree.id,
                name=name,
                groupId=tree.group.id if tree.group != annotation.skeleton else None,
                color=color,
            )
        )

    groups = annotation.skeleton.as_nml_group().children

    volumes = [
        wknml.Volume(
            id=volume.id,
            location=volume._default_zip_name(),
            fallback_layer=volume.fallback_layer_name,
            name=volume.name,
            segments=[
                wknml.Segment(
                    id=segment_id,
                    name=segment_info.name,
                    anchor_position=segment_info.anchor_position,
                    color=segment_info.color,
                )
                for segment_id, segment_info in volume.segments.items()
            ],
            format=str(volume.data_format),
            largest_segment_id=volume.largest_segment_id,
        )
        for volume in annotation._volume_layers
    ]

    meta = [
        wknml.Meta(name=key, content=value)
        for key, value in annotation.metadata.items()
    ]
    if annotation.owner_name is not None:
        meta.append(wknml.Meta(name="username", content=annotation.owner_name))
    if annotation.annotation_id is not None:
        meta.append(wknml.Meta(name="annotationId", content=annotation.annotation_id))

    nml = wknml.Nml(
        meta=meta,
        parameters=nmlParameters,
        trees=trees,
        branchpoints=branchpoints,
        comments=comments,
        groups=groups,
        volumes=volumes,
    )

    return nml


def _extract_nodes_and_edges_from_tree(
    graph: nx.Graph,
) -> Tuple[List[wknml.Node], List[wknml.Edge]]:
    """
    A utility to convert a single [NetworkX graph object](https://networkx.org/) into a list of `NmlNode` objects and `Edge` objects.

    Return
        Tuple[List[NmlNode], List[Edge]]: A tuple containing both all nodes and all edges
    """

    node_nml = [
        wknml.Node(
            id=node.id,
            position=node.position,
            radius=node.radius,
            rotation=node.rotation,
            inVp=node.inVp,
            inMag=node.inMag,
            bitDepth=node.bitDepth,
            interpolation=node.interpolation,
            time=node.time,
        )
        for node in graph.nodes
    ]

    edge_nml = [
        wknml.Edge(source=source.id, target=target.id) for source, target in graph.edges
    ]

    return node_nml, edge_nml



"""This module takes care of uploading datasets to a WEBKNOSSOS server."""

from typing import Any, Optional

import typer
from typing_extensions import Annotated

from ..client import webknossos_context
from ..client._defaults import DEFAULT_WEBKNOSSOS_URL
from ..client._upload_dataset import DEFAULT_SIMULTANEOUS_UPLOADS
from ..dataset import Dataset
from ._utils import parse_path


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            show_default=False,
            help="Path to your local WEBKNOSSOS dataset.",
            parser=parse_path,
        ),
    ],
    webknossos_url: Annotated[
        str,
        typer.Option(
            help="URL to WEBKNOSSOS instance.",
            rich_help_panel="WEBKNOSSOS context",
            envvar="WK_URL",
        ),
    ] = DEFAULT_WEBKNOSSOS_URL,
    token: Annotated[
        Optional[str],
        typer.Option(
            help="Authentication token for WEBKNOSSOS instance "
            "(https://webknossos.org/auth/token).",
            rich_help_panel="WEBKNOSSOS context",
            envvar="WK_TOKEN",
        ),
    ] = None,
    dataset_name: Annotated[
        Optional[str],
        typer.Option(
            help="Alternative name to rename your dataset on upload to WEBKNOSSOS. "
            "(if not provided, current name of dataset is used)",
        ),
    ] = None,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = DEFAULT_SIMULTANEOUS_UPLOADS,
) -> None:
    """Upload a dataset to a WEBKNOSSOS server."""

    with webknossos_context(url=webknossos_url, token=token):
        Dataset.open(dataset_path=source).upload(
            new_dataset_name=dataset_name, jobs=jobs
        )



from typing import NamedTuple, Optional
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import Vector3, enforce_not_null


class Node(NamedTuple):
    id: int
    position: Vector3
    radius: Optional[float] = (
        None  # radius of a node when rendered in wK, unit: nanometers (nm)
    )
    # 3D rotation of the camera when the node was annotated.
    # Mostly relevant for `Flight` mode to resume in the same direction when returning to `Flight` mode.
    rotation: Optional[Vector3] = None
    inVp: Optional[int] = (
        None  # wK UI viewport in which the node was set, `0`: XY plane, `1`: YZ plane. `2`: XY plane, `3`: 3D viewport
    )
    # wK rendering magnification-level when the node was annotated.
    # Lower magnification levels typically indicate a "zoomed-in" workflow resulting in more accurate annotations.
    inMag: Optional[int] = None
    # wK rendering bit-depth when the node was annotated. 4bit (lower data quality) or 8bit (regular quality).
    # Lower quality data rendering might lead to less accurate annotations.
    bitDepth: Optional[int] = None
    # wK rendering interpolation flag when the node was annotated.
    # Interpolated data rendering might lead to less accurate annotations.
    interpolation: Optional[bool] = None
    time: Optional[int] = None  # unix timestamp marking the creation time

    def _dump(self, xf: XmlWriter) -> None:
        attributes = {
            "id": str(self.id),
            "x": str(float(self.position[0])),
            "y": str(float(self.position[1])),
            "z": str(float(self.position[2])),
        }

        if self.radius is not None:
            attributes["radius"] = str(self.radius)

        if self.rotation is not None:
            attributes["rotX"] = str(self.rotation[0])
            attributes["rotY"] = str(self.rotation[1])
            attributes["rotZ"] = str(self.rotation[2])

        if self.inVp is not None:
            attributes["inVp"] = str(self.inVp)

        if self.inMag is not None:
            attributes["inMag"] = str(self.inMag)

        if self.bitDepth is not None:
            attributes["bitDepth"] = str(self.bitDepth)

        if self.interpolation is not None:
            attributes["interpolation"] = str(self.interpolation)

        if self.time is not None:
            attributes["time"] = str(self.time)

        xf.tag("node", attributes)

    @classmethod
    def _parse(cls, nml_node: Element) -> "Node":
        rotation = None
        if nml_node.get("rotX") is not None:
            rotation = (
                float(enforce_not_null(nml_node.get("rotX"))),
                float(enforce_not_null(nml_node.get("rotY"))),
                float(enforce_not_null(nml_node.get("rotZ"))),
            )

        return cls(
            id=int(enforce_not_null(nml_node.get("id"))),
            radius=float(nml_node.get("radius", 0))
            if nml_node.get("radius") is not None
            else None,
            position=(
                float(nml_node.get("x", 0)),
                float(nml_node.get("y", 0)),
                float(nml_node.get("z", 0)),
            ),
            rotation=rotation,
            inVp=int(nml_node.get("inVp", 0))
            if nml_node.get("inVp") is not None
            else None,
            inMag=int(nml_node.get("inMag", 0))
            if nml_node.get("inMag") is not None
            else None,
            bitDepth=int(nml_node.get("bitDepth", 0))
            if nml_node.get("bitDepth") is not None
            else None,
            interpolation=bool(nml_node.get("interpolation"))
            if nml_node.get("interpolation") is not None
            else None,
            time=int(nml_node.get("time", 0))
            if nml_node.get("time") is not None
            else None,
        )



import logging
from pathlib import Path
from typing import Tuple

import wkw

from ..utils import warn_deprecated


def compress_file_job(args: Tuple[Path, Path]) -> None:
    """Deprecated, please use `compress()` on a `MagView` instead."""
    warn_deprecated("compress_file_job", "MagView.compress()")
    source_path, target_path = args
    try:
        target_path.parent.mkdir(parents=True, exist_ok=True)
        wkw.File.compress(str(source_path), str(target_path))

        if not target_path.exists():
            raise Exception("Did not create compressed file {}".format(target_path))

    except Exception as exc:
        logging.error(
            "Compression of '{}' to '{}' failed with {}".format(
                source_path, target_path, exc
            )
        )
        raise exc



from typing import Iterable, List

import attr

from ..client.api_client.models import ApiFolderWithParent


def _get_folder_path(
    folder: ApiFolderWithParent,
    all_folders: Iterable[ApiFolderWithParent],
) -> str:
    if folder.parent is None:
        return folder.name
    else:
        return f"{_get_folder_path(next(f for f in all_folders if f.id == folder.parent), all_folders)}/{folder.name}"


@attr.frozen
class RemoteFolder:
    id: str
    name: str

    @classmethod
    def get_by_id(cls, folder_id: str) -> "RemoteFolder":
        from ..client.context import _get_api_client

        client = _get_api_client(enforce_auth=True)
        folder_tree_response: List[ApiFolderWithParent] = client.folder_tree()

        for folder_info in folder_tree_response:
            if folder_info.id == folder_id:
                return cls(name=folder_info.name, id=folder_info.id)

        raise KeyError(f"Could not find folder {folder_id}.")

    @classmethod
    def get_by_path(cls, path: str) -> "RemoteFolder":
        from ..client.context import _get_api_client

        client = _get_api_client(enforce_auth=True)
        folder_tree_response: List[ApiFolderWithParent] = client.folder_tree()

        for folder_info in folder_tree_response:
            folder_path = _get_folder_path(folder_info, folder_tree_response)
            if folder_path == path:
                return cls(name=folder_info.name, id=folder_info.id)

        raise KeyError(f"Could not find folder {path}.")



import os
import traceback
import warnings
from logging import error, info
from os import getpid
from types import TracebackType
from typing import TYPE_CHECKING, Generator, List, Optional, Type

import numpy as np
import psutil

from webknossos.geometry.nd_bounding_box import NDBoundingBox

from ...geometry import BoundingBox, Vec3Int, Vec3IntLike

if TYPE_CHECKING:
    from ..view import View


def log_memory_consumption(additional_output: str = "") -> None:
    pid = os.getpid()
    process = psutil.Process(pid)
    info(
        "Currently consuming {:.2f} GB of memory ({:.2f} GB still available) "
        "in process {}. {}".format(
            process.memory_info().rss / 1024**3,
            psutil.virtual_memory().available / 1024**3,
            pid,
            additional_output,
        )
    )


class BufferedSliceWriter:
    def __init__(
        self,
        view: "View",
        offset: Optional[Vec3IntLike] = None,
        # json_update_allowed enables the update of the bounding box and rewriting of the properties json.
        # It should be False when parallel access is intended.
        json_update_allowed: bool = True,
        # buffer_size specifies, how many slices should be aggregated until they are flushed.
        buffer_size: int = 32,
        dimension: int = 2,  # z
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        use_logging: bool = False,
    ) -> None:
        """see `View.get_buffered_slice_writer()`"""

        self._view = view
        self._buffer_size = buffer_size
        self._dtype = self._view.get_dtype()
        self._use_logging = use_logging
        self._json_update_allowed = json_update_allowed
        self._bbox: NDBoundingBox
        self._slices_to_write: List[np.ndarray] = []
        self._current_slice: Optional[int] = None
        self._buffer_start_slice: Optional[int] = None

        self.reset_offset(
            offset,
            relative_offset,
            absolute_offset,
            relative_bounding_box,
            absolute_bounding_box,
        )

        assert 0 <= dimension <= 2  # either x (0), y (1) or z (2)
        self.dimension = dimension

        view_chunk_depth = self._view.info.chunk_shape[dimension]
        if (
            self._bbox is not None
            and self._bbox.topleft_xyz[self.dimension] % view_chunk_depth != 0
        ):
            warnings.warn(
                "[WARNING] Using an offset that doesn't align with the datataset's chunk size, "
                + "will slow down the buffered slice writer, because twice as many chunks will be written.",
            )
        if buffer_size >= view_chunk_depth and buffer_size % view_chunk_depth > 0:
            warnings.warn(
                "[WARNING] Using a buffer size that doesn't align with the datataset's chunk size, "
                + "will slow down the buffered slice writer.",
            )

    def _flush_buffer(self) -> None:
        if len(self._slices_to_write) == 0:
            return

        assert (
            len(self._slices_to_write) <= self._buffer_size
        ), "The WKW buffer is larger than the defined batch_size. The buffer should have been flushed earlier. This is probably a bug in the BufferedSliceWriter."

        uniq_dtypes = set(map(lambda _slice: _slice.dtype, self._slices_to_write))
        assert (
            len(uniq_dtypes) == 1
        ), "The buffer of BufferedSliceWriter contains slices with differing dtype."
        assert uniq_dtypes.pop() == self._dtype, (
            "The buffer of BufferedSliceWriter contains slices with a dtype "
            "which differs from the dtype with which the BufferedSliceWriter was instantiated."
        )

        if self._use_logging:
            info(
                "({}) Writing {} slices at position {}.".format(
                    getpid(), len(self._slices_to_write), self._buffer_start_slice
                )
            )
            log_memory_consumption()

        try:
            assert (
                self._buffer_start_slice is not None
            ), "Failed to write buffer: The buffer_start_slice is not set."
            max_width = max(section.shape[-2] for section in self._slices_to_write)
            max_height = max(section.shape[-1] for section in self._slices_to_write)
            channel_count = self._slices_to_write[0].shape[0]
            buffer_depth = min(self._buffer_size, len(self._slices_to_write))
            buffer_start = Vec3Int.zeros().with_replaced(
                self.dimension, self._buffer_start_slice
            )

            bbox = self._bbox.with_size_xyz(
                Vec3Int(max_width, max_height, buffer_depth).moveaxis(
                    -1, self.dimension
                )
            ).offset(buffer_start)

            shard_dimensions = self._view._get_file_dimensions()
            chunk_size = Vec3Int(
                min(shard_dimensions[0], max_width),
                min(shard_dimensions[1], max_height),
                buffer_depth,
            ).moveaxis(-1, self.dimension)
            for chunk_bbox in bbox.chunk(chunk_size):
                info(f"Writing chunk {chunk_bbox}.")

                data = np.zeros(
                    (channel_count, *chunk_bbox.size),
                    dtype=self._slices_to_write[0].dtype,
                )
                section_topleft = Vec3Int(
                    (chunk_bbox.topleft_xyz - bbox.topleft_xyz).moveaxis(
                        self.dimension, -1
                    )
                )
                section_bottomright = Vec3Int(
                    (chunk_bbox.bottomright_xyz - bbox.topleft_xyz).moveaxis(
                        self.dimension, -1
                    )
                )

                z_index = chunk_bbox.index_xyz[self.dimension]

                z = 0
                for section in self._slices_to_write:
                    section_chunk = section[
                        :,
                        section_topleft.x : section_bottomright.x,
                        section_topleft.y : section_bottomright.y,
                    ]
                    # Section chunk includes the axes c, x, y. The remaining axes are added by considering
                    # the length of the bbox. Since the bbox does not contain the channel, we subtract 2
                    # instead of 3.
                    section_chunk = section_chunk[
                        (slice(None), slice(None), slice(None))
                        + tuple(np.newaxis for _ in range(len(bbox) - 2))
                    ]
                    section_chunk = np.moveaxis(
                        section_chunk,
                        [1, 2],
                        bbox.index_xyz[: self.dimension]
                        + bbox.index_xyz[self.dimension + 1 :],
                    )

                    slice_tuple = (slice(None),) + tuple(
                        slice(0, min(size1, size2))
                        for size1, size2 in zip(
                            chunk_bbox.size, section_chunk.shape[1:]
                        )
                    )

                    data[
                        slice_tuple[:z_index]
                        + (slice(z, z + 1),)
                        + slice_tuple[z_index + 1 :]
                    ] = section_chunk

                    z += 1

                self._view.write(
                    data,
                    json_update_allowed=self._json_update_allowed,
                    absolute_bounding_box=chunk_bbox.from_mag_to_mag1(self._view._mag),
                )
                del data

        except Exception as exc:
            error(
                "({}) An exception occurred in BufferedSliceWriter._flush_buffer with {} "
                "slices at position {}. Original error is:\n{}:{}\n\nTraceback:".format(
                    getpid(),
                    len(self._slices_to_write),
                    self._buffer_start_slice,
                    type(exc).__name__,
                    exc,
                )
            )
            traceback.print_tb(exc.__traceback__)
            error("\n")

            raise exc
        finally:
            self._slices_to_write = []

    def _get_slice_generator(self) -> Generator[None, np.ndarray, None]:
        current_slice = 0
        while True:
            data = yield  # Data gets send from the user
            if len(self._slices_to_write) == 0:
                self._buffer_start_slice = current_slice
            if len(data.shape) == 2:
                # The input data might contain channel data or not.
                # Bringing it into the same shape simplifies the code
                data = np.expand_dims(data, axis=0)
            self._slices_to_write.append(data)
            current_slice += 1

            if current_slice % self._buffer_size == 0:
                self._flush_buffer()

    def send(self, value: np.ndarray) -> None:
        self._generator.send(value)

    def reset_offset(
        self,
        offset: Optional[Vec3IntLike] = None,  # deprecated, relative in current mag
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
    ) -> None:
        if self._slices_to_write:
            self._flush_buffer()

        # Reset the generator
        self._generator = self._get_slice_generator()
        next(self._generator)

        if (
            offset is None
            and relative_offset is None
            and absolute_offset is None
            and relative_bounding_box is None
            and absolute_bounding_box is None
        ):
            relative_offset = Vec3Int.zeros()
        if offset is not None:
            warnings.warn(
                "[DEPRECATION] Using offset for a buffered slice writer is deprecated. "
                + "Please use the parameter relative_offset or absolute_offset in Mag(1) instead.",
                DeprecationWarning,
            )

        if offset is not None:
            self._bbox = BoundingBox(
                self._view.bounding_box.topleft_xyz + Vec3Int(offset) * self._view.mag,
                Vec3Int.zeros(),
            )

        if relative_offset is not None:
            self._bbox = BoundingBox(
                self._view.bounding_box.topleft + relative_offset, Vec3Int.zeros()
            )

        if absolute_offset is not None:
            self._bbox = BoundingBox(absolute_offset, Vec3Int.zeros())

        if relative_bounding_box is not None:
            self._bbox = relative_bounding_box.offset(self._view.bounding_box.topleft)

        if absolute_bounding_box is not None:
            self._bbox = absolute_bounding_box

    def __enter__(self) -> "BufferedSliceWriter":
        self._generator = self._get_slice_generator()
        next(self._generator)
        return self

    def __exit__(
        self,
        _type: Optional[Type[BaseException]],
        _value: Optional[BaseException],
        _tb: Optional[TracebackType],
    ) -> None:
        self._flush_buffer()



# ruff: noqa: F401 imported but unused

from .bounding_box import BoundingBox
from .mag import Mag
from .nd_bounding_box import NDBoundingBox
from .vec3_int import Vec3Int, Vec3IntLike
from .vec_int import VecInt, VecIntLike



"""This module takes care of exporting tiff images."""

import logging
from argparse import Namespace
from functools import partial
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Optional, Tuple, Union

import numpy as np
import typer
from PIL import Image
from scipy.ndimage.interpolation import zoom
from typing_extensions import Annotated

from ..dataset import Dataset, MagView, View
from ..dataset.defaults import DEFAULT_CHUNK_SHAPE
from ..geometry import BoundingBox, Mag, Vec3Int
from ..utils import get_executor_for_args, wait_and_ensure_success
from ._utils import (
    DistributionStrategy,
    Vec2Int,
    parse_bbox,
    parse_mag,
    parse_path,
    parse_vec2int,
)


def _make_tiff_name(name: str, slice_index: int) -> str:
    if name is None or name == "":
        return f"{slice_index:06d}.tiff"
    else:
        return f"{name}_{slice_index:06d}.tiff"


def _slice_to_image(data_slice: np.ndarray, downsample: int = 1) -> Image:
    if data_slice.shape[0] == 1:
        # discard greyscale dimension
        data_slice = data_slice.squeeze(axis=0)
        # swap the axis
        data_slice = data_slice.transpose((1, 0))
    else:
        # swap axis and move the channel axis
        data_slice = data_slice.transpose((2, 1, 0))

    if downsample > 1:
        data_slice = zoom(
            data_slice,
            1 / downsample,
            output=data_slice.dtype,
            order=1,
            # this does not mean nearest interpolation,
            # it corresponds to how the borders are treated.
            mode="nearest",
            prefilter=True,
        )
    return Image.fromarray(data_slice)


def export_tiff_slice(
    dest_path: Path,
    name: str,
    tiling_size: Union[None, Tuple[int, int]],
    downsample: int,
    start_slice_index: int,
    view: View,
) -> None:
    tiff_bbox_mag1 = view.bounding_box
    tiff_bbox = tiff_bbox_mag1.in_mag(view.mag)

    if tiling_size is None:
        tiff_data = view.read()
    else:
        padded_tiff_bbox_size = Vec3Int(
            tiling_size[0] * ceil(tiff_bbox.size.x / tiling_size[0]),
            tiling_size[1] * ceil(tiff_bbox.size.y / tiling_size[1]),
            tiff_bbox.size.z,
        )
        tiff_data = view.read()
        padded_tiff_data = np.zeros(
            (tiff_data.shape[0],) + padded_tiff_bbox_size.to_tuple(),
            dtype=tiff_data.dtype,
        )
        padded_tiff_data[
            :, 0 : tiff_data.shape[1], 0 : tiff_data.shape[2], 0 : tiff_data.shape[3]
        ] = tiff_data
        tiff_data = padded_tiff_data
    for slice_index in range(tiff_bbox.size.z):
        slice_name_number = (
            tiff_bbox_mag1.topleft.z + slice_index + 1 - start_slice_index
        )
        if tiling_size is None:
            tiff_file_name = _make_tiff_name(name, slice_name_number)
            tiff_file_path = dest_path / tiff_file_name

            image = _slice_to_image(tiff_data[:, :, :, slice_index], downsample)
            image.save(tiff_file_path)
            logging.debug("Saved slice %s", slice_name_number)

        else:
            for y_tile_index in range(ceil(tiff_bbox.size.y / tiling_size[1])):
                tile_tiff_path = (
                    dest_path / str(slice_name_number) / str(y_tile_index + 1)
                )
                tile_tiff_path.mkdir(parents=True, exist_ok=True)
                for x_tile_index in range(ceil(tiff_bbox.size.x / tiling_size[0])):
                    tile_tiff_filename = f"{x_tile_index + 1}.tiff"
                    tile_image = _slice_to_image(
                        tiff_data[
                            :,
                            x_tile_index * tiling_size[0] : (x_tile_index + 1)
                            * tiling_size[0],
                            y_tile_index * tiling_size[1] : (y_tile_index + 1)
                            * tiling_size[1],
                            slice_index,
                        ],
                        downsample,
                    )

                    tile_image.save(tile_tiff_path / tile_tiff_filename)

            logging.debug("Saved tiles for slice %s", slice_name_number)

    logging.debug("Saved all tiles of bbox %s", tiff_bbox)


def export_tiff_stack(
    mag_view: MagView,
    bbox: BoundingBox,
    destination_path: Path,
    name: str,
    tiling_slice_size: Union[None, Tuple[int, int]],
    batch_size: int,
    downsample: int,
    args: Namespace,
) -> None:
    destination_path.mkdir(parents=True, exist_ok=True)

    with get_executor_for_args(args) as executor:
        view = mag_view.get_view(
            absolute_offset=bbox.topleft, size=bbox.size, read_only=True
        )

        view_chunks = [
            view.get_view(
                relative_offset=(0, 0, z),
                size=(bbox.size.x, bbox.size.y, min(batch_size, bbox.size.z - z)),
                read_only=True,
            )
            for z in range(0, bbox.size.z, batch_size)
        ]

        wait_and_ensure_success(
            executor.map_to_futures(
                partial(
                    export_tiff_slice,
                    destination_path,
                    name,
                    tiling_slice_size,
                    downsample,
                    bbox.topleft.z,
                ),
                view_chunks,
            ),
            executor=executor,
            progress_desc="Exporting tiff files",
        )


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your raw image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    layer_name: Annotated[
        str,
        typer.Option(help="Name of the cubed layer (color or segmentation)"),
    ] = "color",
    bbox: Annotated[
        Optional[BoundingBox],
        typer.Option(
            help="The BoundingBox that should be exported. "
            "If the input data is too small, it will be padded. If it's too large, it will be "
            "cropped. The input format is x,y,z,width,height,depth.",
            parser=parse_bbox,
            metavar="BBOX",
        ),
    ] = None,
    mag: Annotated[
        Mag,
        typer.Option(
            help="Max resolution to be downsampled. "
            "Should be number or minus seperated string (e.g. 2 or 2-2-2).",
            parser=parse_mag,
            metavar="MAG",
        ),
    ] = 1,  # type: ignore
    name: Annotated[str, typer.Option(help="Name of the tiffs.")] = "",
    downsample: Annotated[int, typer.Option(help="Downsample each tiff image.")] = 1,
    tiles_per_dimension: Annotated[
        Optional[Vec2Int],
        typer.Option(
            help="For very large datasets, it is recommended to enable tiling which will ensure "
            "that each slice is exported to multiple images (i.e., tiles). As a parameter you "
            'should provide the amount of tiles per dimension in the form of "x,y".'
            'Also see at "--tile_size" to specify the absolute size of the tiles.',
            parser=parse_vec2int,
        ),
    ] = None,
    tile_size: Annotated[
        Optional[Vec2Int],
        typer.Option(
            help="For very large datasets, it is recommended to enable tiling which will ensure "
            "that each slice is exported to multiple images (i.e., tiles). As a parameter you "
            'should provide the the size of each tile per dimension in the form of "x,y".'
            'Also see at "--tiles_per_dimension" to specify the number of tiles in the dimensions.',
            parser=parse_vec2int,
        ),
    ] = None,
    batch_size: Annotated[
        int, typer.Option(help="Number of sections to buffer per job.")
    ] = DEFAULT_CHUNK_SHAPE.z,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Export your WEBKNOSSOS dataset to TIFF image data."""

    mag_view = Dataset.open(source).get_layer(layer_name).get_mag(mag)

    bbox = BoundingBox.from_ndbbox(mag_view.bounding_box) if bbox is None else bbox

    logging.info("Starting tiff export for bounding box: %s", bbox)
    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )
    used_tile_size = None
    if tiles_per_dimension is not None:
        tile_size = tiles_per_dimension
        logging.info(
            "Using tiling with %d,%d tiles in the dimensions.",
            tile_size[0],
            tile_size[1],
        )
        used_tile_size = (
            ceil(bbox.in_mag(mag_view.mag).size.x / tile_size[0]),
            ceil(bbox.in_mag(mag_view.mag).size.y / tile_size[1]),
        )

    elif tile_size is not None:
        logging.info("Using tiling with the size of %d,%d.", tile_size[0], tile_size[1])
        used_tile_size = tile_size

    export_tiff_stack(
        mag_view=mag_view,
        bbox=bbox,
        destination_path=target,
        name=name,
        tiling_slice_size=used_tile_size,
        batch_size=batch_size,
        downsample=downsample,
        args=executor_args,
    )



from __future__ import division

from collections import namedtuple
from functools import partial
from pathlib import Path
from threading import Lock
from typing import Callable, List, Optional

from .util import CallbackDispatcher

FileChunk = namedtuple("FileChunk", ["index", "size", "read"])


def build_chunks(
    read_bytes: Callable[[int, int], bytes], file_size: int, chunk_size: int
) -> List[FileChunk]:
    chunks = []

    index = 0
    start = 0

    while start < file_size:
        end = min(start + chunk_size, file_size)
        size = end - start

        chunk = FileChunk(index, size, partial(read_bytes, start, size))
        chunks.append(chunk)

        index += 1
        start += chunk_size

    return chunks


class ResumableFile:
    """A file to be uploaded in a resumable session.

    Parameters
    ----------
    path : str or pathlib.Path
        The path of the file
    relative_path : pathlib.Path or None
        The relative path of the file
    chunk_size : int
        The size, in bytes, of chunks uploaded in a single request

    Attributes
    ----------
    completed : resumable.util.CallbackDispatcher
        Triggered when all chunks of the file have been uploaded
    chunk_completed : resumable.util.CallbackDispatcher
        Triggered when a chunks of the file has been uploaded, passing the
        chunk
    """

    def __init__(
        self,
        path: Path,
        relative_path: Optional[Path],
        chunk_size: int,
        generate_unique_identifier: Callable[[Path, Path], str],
    ) -> None:
        self.path = path
        if relative_path is not None:
            self.relative_path = relative_path
        else:
            self.relative_path = path
        self.unique_identifier = generate_unique_identifier(
            self.path, self.relative_path
        )
        self.chunk_size = chunk_size
        self.size = self.path.stat().st_size

        self._fp = open(self.path, "rb")
        self._fp_lock = Lock()

        self.chunks = build_chunks(self._read_bytes, self.size, chunk_size)
        self._chunk_done = {chunk: False for chunk in self.chunks}

        self.completed = CallbackDispatcher()
        self.chunk_completed = CallbackDispatcher()

    def close(self) -> None:
        """Close the file."""
        self._fp.close()

    def _read_bytes(self, start: int, num_bytes: int) -> bytes:
        """Read a byte range from the file."""
        with self._fp_lock:
            self._fp.seek(start)
            return self._fp.read(num_bytes)

    @property
    def is_completed(self) -> bool:
        """Indicates if all chunks of this file have been uploaded."""
        return all(self._chunk_done.values())

    @property
    def fraction_completed(self) -> float:
        """The fraction of the file that has been completed."""
        return sum(self._chunk_done.values()) / len(self.chunks)

    def mark_chunk_completed(self, chunk: FileChunk) -> None:
        """Mark a chunk of this file as having been successfully uploaded.

        If all chunks have been completed, this will trigger the `completed`
        callback of this file.

        Parameters
        ----------
        chunk : resumable.chunk.FileChunk
            The chunk to mark as completed
        """
        self._chunk_done[chunk] = True
        if self.is_completed:
            self.completed.trigger()
            self.close()
        self.chunk_completed.trigger(chunk)



import argparse
import calendar
import functools
import json
import logging
import sys
import time
import warnings
from concurrent.futures import Future
from contextlib import nullcontext
from datetime import datetime
from inspect import getframeinfo, stack
from multiprocessing import cpu_count
from os.path import relpath
from pathlib import Path
from shutil import copyfileobj
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
)

import numpy as np
import rich
from cluster_tools import Executor, get_executor
from rich.progress import Progress
from upath import UPath

times = {}


def time_start(identifier: str) -> None:
    times[identifier] = time.time()
    logging.debug("{} started".format(identifier))


def time_stop(identifier: str) -> None:
    _time = times.pop(identifier)
    logging.debug("{} took {:.8f}s".format(identifier, time.time() - _time))


def get_executor_for_args(
    args: Optional[argparse.Namespace],
    executor: Optional[Executor] = None,
) -> ContextManager[Executor]:
    if executor is not None:
        return nullcontext(enter_result=executor)

    if args is None:
        # For backwards compatibility with code from other packages
        # we allow args to be None. In this case we are defaulting
        # to these values:
        jobs = cpu_count()
        executor = get_executor("multiprocessing", max_workers=jobs)
        logging.info("Using pool of {} workers.".format(jobs))
    elif args.distribution_strategy == "multiprocessing":
        # Also accept "processes" instead of job to be compatible with segmentation-tools.
        # In the long run, the args should be unified and provided by the clustertools.
        if "jobs" in args:
            jobs = args.jobs
        elif "processes" in args:
            jobs = args.processes
        else:
            jobs = cpu_count()

        executor = get_executor("multiprocessing", max_workers=jobs)
        logging.info("Using pool of {} workers.".format(jobs))
    elif args.distribution_strategy in ("slurm", "kubernetes"):
        if args.job_resources is None:
            resources_example = (
                '{"mem": "1G"}'
                if args.distribution_strategy == "slurm"
                else '{"memory": "1G"}'
            )
            raise argparse.ArgumentTypeError(
                f"Job resources (--job_resources) has to be provided when using {args.distribution_strategy} as distribution strategy. Example: --job_resources='{resources_example}'"
            )

        executor = get_executor(
            args.distribution_strategy,
            debug=True,
            keep_logs=True,
            job_resources=json.loads(args.job_resources),
        )
        logging.info(f"Using {args.distribution_strategy} cluster.")
    elif args.distribution_strategy == "debug_sequential":
        executor = get_executor(
            args.distribution_strategy,
            debug=True,
            keep_logs=True,
        )
    else:
        logging.error(
            "Unknown distribution strategy: {}".format(args.distribution_strategy)
        )

    return executor


F = Callable[..., Any]


def named_partial(func: F, *args: Any, **kwargs: Any) -> F:
    # Propagate __name__ and __doc__ attributes to partial function
    partial_func = functools.partial(func, *args, **kwargs)
    functools.update_wrapper(partial_func, func)
    if hasattr(func, "__annotations__"):
        # Generic types cannot be pickled in Python <= 3.6, see https://github.com/python/typing/issues/511
        partial_func.__annotations__ = {}
    return partial_func


def wait_and_ensure_success(
    futures: List[Future],
    executor: Executor,
    progress_desc: Optional[str] = None,
) -> List[Any]:
    """Waits for all futures to complete and raises an exception
    as soon as a future resolves with an error."""

    results = []
    if progress_desc is None:
        for fut in executor.as_completed(futures):
            results.append(fut.result())  #  noqa: PERF401 Use a list comprehension to create a transformed list
    else:
        with get_rich_progress() as progress:
            task = progress.add_task(progress_desc, total=len(futures))
            for fut in executor.as_completed(futures):
                results.append(fut.result())
                progress.update(task, advance=1)
    return results


def snake_to_camel_case(snake_case_name: str) -> str:
    parts = snake_case_name.split("_")
    return parts[0] + "".join(part.title() for part in parts[1:])


def get_chunks(arr: List[Any], chunk_size: int) -> Iterable[List[Any]]:
    for i in range(0, len(arr), chunk_size):
        yield arr[i : i + chunk_size]


def time_since_epoch_in_ms() -> int:
    d = datetime.utcnow()
    unixtime = calendar.timegm(d.utctimetuple())
    return unixtime * 1000


def copy_directory_with_symlinks(
    src_path: Path,
    dst_path: Path,
    ignore: Iterable[str] = tuple(),
    make_relative: bool = False,
) -> None:
    """
    Links all directories in src_path / dir_name to dst_path / dir_name.
    """
    assert is_fs_path(src_path), f"Cannot create symlink with remote paths {src_path}."
    assert is_fs_path(dst_path), f"Cannot create symlink with remote paths {dst_path}."

    for item in src_path.iterdir():
        if item.name not in ignore:
            symlink_path = dst_path / item.name
            if make_relative:
                rel_or_abspath = Path(relpath(item, symlink_path.parent))
            else:
                rel_or_abspath = item.resolve()
            symlink_path.symlink_to(rel_or_abspath)


def setup_warnings() -> None:
    warnings.filterwarnings("default", category=DeprecationWarning, module="webknossos")


def setup_logging(args: argparse.Namespace) -> None:
    log_path = Path(f"./logs/cuber_{time.strftime('%Y-%m-%d_%H%M%S')}.txt")

    console_log_level = logging.DEBUG if args.verbose else logging.INFO
    file_log_level = logging.DEBUG

    logging_formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")

    # Always set the global log level to the more verbose of console_log_level and
    # file_log_level to allow to log with different log levels to console and files.
    root_logger = logging.getLogger()
    root_logger.setLevel(min(console_log_level, file_log_level))

    console = logging.StreamHandler(sys.stdout)
    console.setLevel(console_log_level)
    console.setFormatter(logging_formatter)
    root_logger.addHandler(console)

    log_path.parent.mkdir(parents=True, exist_ok=True)
    file_handler = logging.FileHandler(log_path, mode="w", encoding="UTF-8")
    file_handler.setLevel(file_log_level)
    file_handler.setFormatter(logging_formatter)
    root_logger.addHandler(file_handler)


def add_verbose_flag(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--verbose", help="Verbose output", dest="verbose", action="store_true"
    )

    parser.set_defaults(verbose=False)


def get_rich_progress() -> Progress:
    return Progress(
        "[progress.description]{task.description:<20}",
        rich.progress.BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        rich.progress.TimeElapsedColumn(),
        "|",
        rich.progress.TimeRemainingColumn(),
    )


def warn_deprecated(deprecated_item: str, alternative_item: str) -> None:
    caller = getframeinfo(stack()[2][0])
    warnings.warn(
        f"[DEPRECATION] `{deprecated_item}` is deprecated, please use `{alternative_item}` instead (see {caller.filename}:{caller.lineno})",
        DeprecationWarning,
        stacklevel=2,
    )


def is_fs_path(path: Path) -> bool:
    from upath.implementations.local import PosixUPath, WindowsUPath

    return not isinstance(path, UPath) or isinstance(path, (PosixUPath, WindowsUPath))


def strip_trailing_slash(path: Path) -> Path:
    if isinstance(path, UPath):
        return UPath(
            str(path).rstrip("/"),
            **path._kwargs.copy(),
        )
    else:
        return Path(str(path).rstrip("/"))


def rmtree(path: Path) -> None:
    def _walk(path: Path) -> Iterator[Path]:
        if path.exists():
            if path.is_dir() and not path.is_symlink():
                for p in path.iterdir():
                    yield from _walk(p)
            yield path

    for sub_path in _walk(path):
        try:
            if sub_path.is_file() or sub_path.is_symlink():
                sub_path.unlink()
            elif sub_path.is_dir():
                sub_path.rmdir()
        except FileNotFoundError:  # noqa:  PERF203 `try`-`except` within a loop incurs performance overhead
            # Some implementations `UPath` do not have explicit directory representations
            # Therefore, directories only exist, if they have files. Consequently, when
            # all files have been deleted, the directory does not exist anymore.
            pass


def copytree(in_path: Path, out_path: Path) -> None:
    def _walk(path: Path, base_path: Path) -> Iterator[Tuple[Path, Path]]:
        yield (path, path.relative_to(base_path))
        if path.is_dir():
            for p in path.iterdir():
                yield from _walk(p, base_path)

    for in_sub_path, sub_path in _walk(in_path, in_path):
        if in_sub_path.is_dir():
            (out_path / sub_path).mkdir(parents=True, exist_ok=True)
        else:
            with (in_path / sub_path).open("rb") as in_file, (out_path / sub_path).open(
                "wb"
            ) as out_file:
                copyfileobj(in_file, out_file)


K = TypeVar("K")  # key
V = TypeVar("V")  # value
C = TypeVar("C")  # cache


class LazyReadOnlyDict(Mapping[K, V]):
    def __init__(self, entries: Dict[K, C], func: Callable[[C], V]) -> None:
        self.entries = entries
        self.func = func

    def __getitem__(self, key: K) -> V:
        return self.func(self.entries[key])

    def __iter__(self) -> Iterator[K]:
        for key in self.entries:
            yield key

    def __len__(self) -> int:
        return len(self.entries)


class NDArrayLike(Protocol):
    def __getitem__(self, selection: Tuple[slice, ...]) -> np.ndarray: ...

    def __setitem__(self, selection: Tuple[slice, ...], value: np.ndarray) -> None: ...

    @property
    def shape(self) -> Tuple[int, ...]: ...

    @property
    def ndim(self) -> int: ...

    @property
    def dtype(self) -> np.dtype: ...



from typing import NamedTuple
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import enforce_not_null


class Edge(NamedTuple):
    source: int
    target: int

    def _dump(self, xf: XmlWriter) -> None:
        xf.tag("edge", {"source": str(self.source), "target": str(self.target)})

    @classmethod
    def _parse(cls, nml_edge: Element) -> "Edge":
        return cls(
            source=int(enforce_not_null(nml_edge.get("source"))),
            target=int(enforce_not_null(nml_edge.get("target"))),
        )



import re
from operator import add, floordiv, mod, mul, sub
from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar, Union, cast

import numpy as np

_VALUE_ERROR = "VecInt can be instantiated with int values `VecInt(1,2,3,4) or with `VecIntLike` object `VecInt([1,2,3,4])."

_T = TypeVar("_T", bound="VecInt")


class VecInt(tuple):
    """
    The VecInt class is designed to represent a vector of integers. This class is a subclass of the built-in tuple class, and it extends the functionality of tuples by providing additional methods and operations.

    One of the key features of the VecInt class is that it allows for the storage of axis names along with their corresponding values.

    Here is a brief example demonstrating how to use the VecInt class:

    ```python
    from webknossos import VecInt

    # Creating a VecInt instance with 4 elements and axes x, y, z, t:
    vector_1 = VecInt(1, 2, 3, 4, axes=("x", "y", "z", "t"))
    # Alternative ways to create the same VecInt instance:
    vector_1 = VecInt([1, 2, 3, 4], axes=("x", "y", "z", "t"))
    vector_1 = VecInt(x=1, y=2, z=3, t=4)

    # Creating a VecInt instance with all elements set to 1 and axes x, y, z, t:
    vector_2 = VecInt.full(1, axes=("x", "y", "z", "t"))
    # Asserting that all elements in vector_2 are equal to 1:
    assert vector_2[0] == vector_2[1] == vector_2[2] == vector_2[3]

    # Demonstrating the addition operation between two VecInt instances:
    assert vector_1 + vector_2 == VecInt(2, 3, 4, 5)
    ```
    """

    axes: Tuple[str, ...]
    _x_pos: Optional[int]
    _y_pos: Optional[int]
    _z_pos: Optional[int]

    def __new__(
        cls,
        *args: Union["VecIntLike", Iterable[str], int],
        axes: Optional[Iterable[str]] = None,
        **kwargs: int,
    ) -> "VecInt":
        as_tuple: Optional[Tuple[int, ...]] = None

        if args:
            if isinstance(args[0], VecInt):
                return args[0]
            if isinstance(args[0], np.ndarray):
                assert np.count_nonzero(args[0] % 1) == 0, _VALUE_ERROR
            if isinstance(args[0], str):
                return cls.from_str(args[0])
            if isinstance(args[0], Iterable):
                as_tuple = tuple(int(item) for item in args[0])
                if args[1:] and isinstance(args[1], Iterable):
                    assert all(isinstance(arg, str) for arg in args[1]), _VALUE_ERROR
                    axes = tuple(args[1])  # type: ignore
            elif isinstance(args, Iterable):
                as_tuple = tuple(int(arg) for arg in args)  # type: ignore
            else:
                raise ValueError(_VALUE_ERROR)
            assert axes is not None, _VALUE_ERROR
        else:
            assert kwargs, _VALUE_ERROR
            assert axes is None, _VALUE_ERROR
            as_tuple = tuple(kwargs.values())

        assert as_tuple is not None, _VALUE_ERROR

        self = super().__new__(cls, cast(Iterable, as_tuple))
        # self.axes is set in __new__ instead of __init__ so that pickling/unpickling
        # works without problems. As long as the deserialization of a tree instance
        # is not finished, the object is only half-initialized. Since self.axes
        # is needed after deepcopy, an error would be raised otherwise.
        # Also see:
        # https://stackoverflow.com/questions/46283738/attributeerror-when-using-python-deepcopy
        self.axes = tuple(axes or kwargs.keys())
        self._x_pos = self.axes.index("x") if "x" in self.axes else None
        self._y_pos = self.axes.index("y") if "y" in self.axes else None
        self._z_pos = self.axes.index("z") if "z" in self.axes else None

        return self

    def __getnewargs__(self) -> Tuple[Tuple[int, ...], Tuple[str, ...]]:
        return (self.to_tuple(), self.axes)

    @property
    def x(self) -> int:
        """
        Returns the x component of the vector.
        """
        if self._x_pos is not None:
            return self[self._x_pos]

        raise ValueError("The vector does not have an x component.")

    @property
    def y(self) -> int:
        """
        Returns the y component of the vector.
        """
        if self._y_pos is not None:
            return self[self._y_pos]

        raise ValueError("The vector does not have an y component.")

    @property
    def z(self) -> int:
        """
        Returns the z component of the vector.
        """
        if self._z_pos is not None:
            return self[self._z_pos]

        raise ValueError("The vector does not have an z component.")

    @staticmethod
    def from_str(string: str) -> "VecInt":
        """
        Returns a new ND Vector from a string representation.

        Args:
        - string (str): The string representation of the vector.

        Returns:
        - VecInt: The new vector.
        """
        return VecInt(tuple(map(int, re.findall(r"\d+", string))))

    def with_replaced(self: _T, index: int, new_element: int) -> _T:
        """Returns a new ND Vector with a replaced element at a given index."""

        return self.__class__(
            *self[:index], new_element, *self[index + 1 :], axes=self.axes
        )

    def to_np(self) -> np.ndarray:
        """
        Returns the vector as a numpy array.
        """
        return np.array(self)

    def to_list(self) -> List[int]:
        """
        Returns the vector as a list.
        """
        return list(self)

    def to_tuple(self) -> Tuple[int, ...]:
        """
        Returns the vector as a tuple.
        """
        return tuple(self)

    def contains(self, needle: int) -> bool:
        """
        Checks if the vector contains a given element.
        """
        return any(element == needle for element in self)

    def is_positive(self, strictly_positive: bool = False) -> bool:
        """
        Checks if all elements in the vector are positive.

        Args:
        - strictly_positive (bool): If True, checks if all elements are strictly positive.

        Returns:
        - bool: True if all elements are positive, False otherwise.
        """
        if strictly_positive:
            return all(i > 0 for i in self)

        return all(i >= 0 for i in self)

    def is_uniform(self) -> bool:
        """
        Checks if all elements in the vector are the same.
        """
        first = self[0]
        return all(element == first for element in self)

    def _element_wise(
        self: _T, other: Union[int, "VecIntLike"], fn: Callable[[int, Any], int]
    ) -> _T:
        if isinstance(other, int):
            other_imported = VecInt.full(other, axes=self.axes)
        else:
            other_imported = VecInt(other, axes=self.axes)
            assert len(other_imported) == len(
                self
            ), f"{other} and {self} are not equally shaped."
        return self.__class__(
            **{
                axis: fn(self[i], other_imported[i]) for i, axis in enumerate(self.axes)
            },
            axes=None,
        )

    # Note: When adding regular tuples the first tuple is extended with the second tuple.
    # For VecInt we want to add the elements at the same index.
    # Do not add VecInt to plain tuple! Hence the type:ignore)
    def __add__(self: _T, other: Union[int, "VecIntLike"]) -> _T:  # type: ignore[override]
        return self._element_wise(other, add)

    def __sub__(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        return self._element_wise(other, sub)

    # Note: When multiplying regular tuples with an int those are repeated,
    # which is a different behavior in the superclass! Hence the type:ignore.
    def __mul__(self: _T, other: Union[int, "VecIntLike"]) -> _T:  # type: ignore[override]
        return self._element_wise(other, mul)

    def __floordiv__(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        return self._element_wise(other, floordiv)

    def __mod__(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        return self._element_wise(other, mod)

    def __neg__(self: _T) -> _T:
        return self.__class__((-elem for elem in self), axes=self.axes)

    def ceildiv(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        """
        Returns a new VecInt with the ceil division of each element by the other.
        """
        return (self + other - 1) // other

    def pairmax(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        """
        Returns a new VecInt with the maximum of each pair of elements from the two vectors.
        """
        return self._element_wise(other, max)

    def pairmin(self: _T, other: Union[int, "VecIntLike"]) -> _T:
        """
        Returns a new VecInt with the minimum of each pair of elements from the two vectors.
        """
        return self._element_wise(other, min)

    def prod(self) -> int:
        """
        Returns the product of all elements in the vector.
        """
        return int(np.prod(self.to_np()))

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({','.join((str(element) for element in self))})"
        )

    def add_or_none(self: _T, other: Optional["VecInt"]) -> Optional[_T]:
        """
        Adds two VecInts or returns None if the other is None.

        Args:
        - other (Optional[VecInt]): The other vector to add.

        Returns:
        - Optional[VecInt]: The sum of the two vectors or None if the other is None.
        """
        return None if other is None else self + other

    def moveaxis(
        self: _T, source: Union[int, List[int]], target: Union[int, List[int]]
    ) -> _T:
        """
        Allows to move one element at index `source` to another index `target`. Similar to
        np.moveaxis, this is *not* a swap operation but instead it moves the specified
        source so that the other elements move when necessary.

        Args:
        - source (Union[int, List[int]]): The index of the element to move.
        - target (Union[int, List[int]]): The index where the element should be moved to.

        Returns:
        - VecInt: A new vector with the moved element.
        """

        # Piggy-back on np.moveaxis by creating an auxiliary array where the indices 0, 1 and
        # 2 appear in the shape.
        indices = np.moveaxis(
            np.zeros(tuple(i for i in range(len(self)))), source, target
        ).shape
        arr = self.to_np()[np.array(indices)]
        axes = np.array(self.axes)[np.array(indices)]
        return self.__class__(arr, axes=axes)

    @classmethod
    def zeros(cls, axes: Tuple[str, ...]) -> "VecInt":
        """
        Returns a new ND Vector with all elements set to 0.

        Args:
        - axes (Tuple[str, ...]): The axes of the vector.

        Returns:
        - VecInt: The new vector.
        """
        return cls((0 for _ in range(len(axes))), axes=axes)

    @classmethod
    def ones(cls, axes: Tuple[str, ...]) -> "VecInt":
        """
        Returns a new ND Vector with all elements set to 1.

        Args:
        - axes (Tuple[str, ...]): The axes of the vector.

        Returns:
        - VecInt: The new vector.
        """
        return cls((1 for _ in range(len(axes))), axes=axes)

    @classmethod
    def full(cls, an_int: int, axes: Tuple[str, ...]) -> "VecInt":
        """
        Returns a new ND Vector with all elements set to the same value.

        Args:
        - an_int (int): The value to set all elements to.
        - axes (Tuple[str, ...]): The axes of the vector.

        Returns:
        - VecInt: The new vector.
        """
        return cls((an_int for _ in range(len(axes))), axes=axes)


VecIntLike = Union[VecInt, Tuple[int, ...], np.ndarray, Iterable[int]]



import re
from typing import Iterable, Optional, Tuple, Union, cast

import numpy as np

from .vec_int import VecInt

_VALUE_ERROR = "Vector components must be three integers or a Vec3IntLike object."


class Vec3Int(VecInt):
    def __new__(
        cls,
        *args: Union["Vec3IntLike", Iterable[str], int],
        axes: Optional[Iterable[str]] = ("x", "y", "z"),
        **kwargs: int,
    ) -> "Vec3Int":
        """
        Class to represent a 3D vector. Inherits from tuple and provides useful
        methods and operations on top.

        A small usage example:

        ```python
        from webknossos import Vec3Int

        vector_1 = Vec3Int(1, 2, 3)
        vector_2 = Vec3Int.full(1)
        assert vector_2.x == vector_2.y == vector_2.y

        assert vector_1 + vector_2 == (2, 3, 4)
        ```
        """

        if args:
            if isinstance(args[0], Vec3Int):
                return args[0]

            assert axes is not None, _VALUE_ERROR

            if isinstance(args[0], Iterable):
                self = super().__new__(cls, *args[0], axes=("x", "y", "z"))
                assert self is not None and len(self) == 3, _VALUE_ERROR

                return cast(Vec3Int, self)

            assert len(args) == 3 and len(tuple(axes)) == 3, _VALUE_ERROR
            assert kwargs is None or len(kwargs) == 0, _VALUE_ERROR
            assert "x" in axes and "y" in axes and "z" in axes, _VALUE_ERROR
            values, _ = zip(*sorted(zip(args, axes), key=lambda x: x[1]))
        else:
            assert "x" in kwargs and "y" in kwargs and "z" in kwargs, _VALUE_ERROR
            assert len(kwargs) == 3, _VALUE_ERROR
            values = kwargs["x"], kwargs["y"], kwargs["z"]

        self = super().__new__(cls, *values, axes=("x", "y", "z"))
        self.axes = ("x", "y", "z")

        assert self is not None and len(self) == 3, _VALUE_ERROR

        return cast(Vec3Int, self)

    def __getnewargs__(self) -> Tuple[Tuple[int, ...], Tuple[str, ...]]:
        return (self.to_tuple(), self.axes)

    @property
    def x(self) -> int:
        return self[0]

    @property
    def y(self) -> int:
        return self[1]

    @property
    def z(self) -> int:
        return self[2]

    def with_x(self, new_x: int) -> "Vec3Int":
        return Vec3Int.from_xyz(new_x, self.y, self.z)

    def with_y(self, new_y: int) -> "Vec3Int":
        return Vec3Int.from_xyz(self.x, new_y, self.z)

    def with_z(self, new_z: int) -> "Vec3Int":
        return Vec3Int.from_xyz(self.x, self.y, new_z)

    def to_tuple(self) -> Tuple[int, int, int]:
        return (self.x, self.y, self.z)

    @staticmethod
    def from_xyz(x: int, y: int, z: int) -> "Vec3Int":
        """Use Vec3Int.from_xyz for fast construction."""

        # By calling __new__ of tuple directly, we circumvent
        # the tolerant (and potentially) slow Vec3Int.__new__ method.
        vec3int = tuple.__new__(Vec3Int, (x, y, z))
        vec3int.axes = ("x", "y", "z")
        return vec3int

    @staticmethod
    def from_vec3_float(vec: Tuple[float, float, float]) -> "Vec3Int":
        return Vec3Int(int(vec[0]), int(vec[1]), int(vec[2]))

    @staticmethod
    def from_vec_or_int(vec_or_int: Union["Vec3IntLike", int]) -> "Vec3Int":
        if isinstance(vec_or_int, int):
            return Vec3Int.full(vec_or_int)

        return Vec3Int(vec_or_int)

    @staticmethod
    def from_str(string: str) -> "Vec3Int":
        if re.match(r"\(\d+,\d+,\d+\)", string):
            return Vec3Int(tuple(map(int, re.findall(r"\d+", string))))

        return Vec3Int.full(int(string))

    @classmethod
    def zeros(cls, _axes: Tuple[str, ...] = ("x", "y", "z")) -> "Vec3Int":
        return cls(0, 0, 0)

    @classmethod
    def ones(cls, _axes: Tuple[str, ...] = ("x", "y", "z")) -> "Vec3Int":
        return cls(1, 1, 1)

    @classmethod
    def full(cls, an_int: int, _axes: Tuple[str, ...] = ("x", "y", "z")) -> "Vec3Int":
        return cls(an_int, an_int, an_int)


Vec3IntLike = Union[Vec3Int, Tuple[int, int, int], np.ndarray, Iterable[int]]



"""This module merges a volume annotation layer with its fallback layer."""

import logging
from argparse import Namespace
from multiprocessing import cpu_count
from typing import Any, List, Optional, Tuple

import typer
from cluster_tools import Executor
from rich.progress import track
from typing_extensions import Annotated

from ..annotation import Annotation
from ..dataset import Dataset, MagView
from ..geometry import BoundingBox, Mag
from ..utils import get_executor_for_args
from ._utils import DistributionStrategy, parse_path


def main(
    *,
    target: Annotated[
        Any,
        typer.Argument(
            help="Path to your WEBKNOSSOS output dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    source_annotation: Annotated[
        Any,
        typer.Argument(
            help="Path to your WEBKNOSSOS zip annotation",
            show_default=False,
            parser=parse_path,
        ),
    ],
    dataset_directory: Annotated[
        Any,
        typer.Argument(
            help="Path to your WEBKNOSSOS dataset folder.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    volume_layer_name: Annotated[
        Optional[str],
        typer.Option(help="Name of the volume layer to merge with fallback layer."),
    ] = None,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Merges a given WEBKNOSSOS annotation."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    annotation = Annotation.load(source_annotation)
    annotation_volumes = list(annotation.get_volume_layer_names())

    output_dataset = Dataset(
        target,
        voxel_size=annotation.voxel_size,
    )
    assert len(annotation_volumes) > 0, "Annotation does not contain any volume layers!"

    if volume_layer_name is not None:
        assert (
            volume_layer_name in annotation_volumes
        ), f'Volume layer name "{volume_layer_name}" not found in annotation'
    else:
        assert (
            len(annotation_volumes) == 1
        ), "Volume layer name was not provided and more than one volume layer found in annotation"
        volume_layer_name = annotation_volumes[0]

    volume_layer = annotation._get_volume_layer(volume_layer_name=volume_layer_name)
    fallback_layer_name = volume_layer.fallback_layer_name

    if fallback_layer_name is None:
        logging.info("No fallback layer found, save annotation as dataset.")
        annotation.export_volume_layer_to_dataset(output_dataset)

    else:
        fallback_dataset_path = dataset_directory / annotation.dataset_name
        fallback_layer = Dataset.open(fallback_dataset_path).get_layer(
            fallback_layer_name
        )

        if volume_layer.zip is None:
            logging.info("No volume annotation found. Copy fallback layer.")
            with get_executor_for_args(args=executor_args) as executor:
                output_dataset.add_copy_layer(
                    fallback_layer, compress=True, executor=executor
                )

        else:
            tmp_annotation_layer_name = f"{annotation.name}-TMP"
            logging.info(
                f"Unpack annotation layer {volume_layer_name} temporarily in {output_dataset.name} as {tmp_annotation_layer_name}"
            )
            # NOTE(erjel): Cannot use "temporary_volume_layer_copy" here, since tmp folders
            # might not be accessible from slurm compute nodes.
            input_annotation_layer = annotation.export_volume_layer_to_dataset(
                output_dataset,
                layer_name=tmp_annotation_layer_name,
                volume_layer_name=volume_layer_name,
            )

            input_annotation_mag = input_annotation_layer.get_finest_mag()
            fallback_mag = fallback_layer.get_mag(input_annotation_mag.mag)

            logging.info(f"Create layer {fallback_layer.name} in {output_dataset.path}")
            output_layer = output_dataset.add_layer_like(
                fallback_layer, fallback_layer.name
            )

            with get_executor_for_args(args=executor_args) as executor:
                logging.info(
                    f"Copy Mag {fallback_mag.mag} from {fallback_layer.path} to {output_layer.path}"
                )
                output_mag = output_layer.add_copy_mag(
                    fallback_mag,
                    compress=True,
                    executor=executor,
                )

                merge_mags(output_mag, input_annotation_mag, executor)

            logging.info("Delete temporary annotation layer")
            output_dataset.delete_layer(tmp_annotation_layer_name)
            logging.info("Done.")


def merge_mags(
    output_mag: MagView,
    input_annotation_mag: MagView,
    executor: Executor,
) -> None:
    assert all(
        input_annotation_mag.info.chunks_per_shard.to_np() == 1
    ), "volume annotation must have file_len=1"
    assert (
        input_annotation_mag.info.voxel_type == output_mag.info.voxel_type
    ), "Volume annotation must have same dtype as fallback layer"
    assert (
        input_annotation_mag.mag == output_mag.mag
    ), f"Volume annotation mag {input_annotation_mag.mag} must match the fallback layer mag {output_mag.mag}"

    logging.info("Scan disk for annotation shards.")
    bboxes = list(bbox for bbox in input_annotation_mag.get_bounding_boxes_on_disk())

    logging.info(f"Grouping {len(bboxes)} bboxes according to output shards.")
    shards_with_bboxes = BoundingBox.group_boxes_with_aligned_mag(
        bboxes, Mag(output_mag.info.shard_shape * output_mag.mag)
    )

    args = [
        (input_annotation_mag, output_mag, shard, bboxes)
        for shard, bboxes in shards_with_bboxes.items()
    ]

    logging.info(f"Merging {len(args)} shards.")
    executor.map(merge_chunk, args)


def merge_chunk(args: Tuple[MagView, MagView, BoundingBox, List[BoundingBox]]) -> None:
    mag_in, mag_out, shard, bboxes = args
    data_buffer = mag_out.read(absolute_bounding_box=shard)[0]

    for bbox in track(bboxes, description="Processing..."):
        read_data = mag_in.read(absolute_bounding_box=bbox)[0]
        data_buffer[bbox.offset(-shard.topleft).in_mag(mag_in.mag).to_slices()] = (
            read_data
        )

    mag_out.write(data_buffer, absolute_offset=shard.topleft)



import re
import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from os.path import relpath
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import numcodecs
import numpy as np
import wkw
import zarr
from upath import UPath
from zarr.storage import FSStore

from ..geometry import BoundingBox, NDBoundingBox, Vec3Int, VecInt
from ..utils import is_fs_path, warn_deprecated
from .data_format import DataFormat

if TYPE_CHECKING:
    import zarrita
    import zarrita.codecs


def _is_power_of_two(num: int) -> bool:
    return num & (num - 1) == 0


def _fsstore_from_path(path: Path, mode: str = "a") -> FSStore:
    storage_options = {}
    if isinstance(path, UPath) and not is_fs_path(path):
        storage_options = path._kwargs.copy()
        storage_options.pop("_url", None)
        return FSStore(url=str(path), mode=mode, **storage_options)

    return FSStore(url=str(path), mode=mode, **storage_options)


@contextmanager
def _blosc_disable_threading() -> Iterator[None]:
    old_value = numcodecs.blosc.use_threads

    # See https://zarr.readthedocs.io/en/stable/tutorial.html#configuring-blosc
    numcodecs.blosc.use_threads = False
    try:
        yield
    finally:
        numcodecs.blosc.use_threads = old_value


class ArrayException(Exception):
    pass


@dataclass
class ArrayInfo:
    data_format: DataFormat
    num_channels: int
    voxel_type: np.dtype
    chunk_shape: Vec3Int
    chunks_per_shard: Vec3Int
    shape: VecInt = VecInt(c=1, x=1, y=1, z=1)
    dimension_names: Tuple[str, ...] = ("c", "x", "y", "z")
    axis_order: VecInt = VecInt(c=3, x=2, y=1, z=0)
    compression_mode: bool = False

    @property
    def shard_size(self) -> Vec3Int:
        warn_deprecated("shard_size", "shard_shape")
        return self.shard_shape

    @property
    def shard_shape(self) -> Vec3Int:
        return self.chunk_shape * self.chunks_per_shard


class BaseArray(ABC):
    data_format = DataFormat.WKW

    _path: Path

    def __init__(self, path: Path):
        self._path = path

    @property
    @abstractmethod
    def info(self) -> ArrayInfo:
        pass

    @classmethod
    @abstractmethod
    def open(_cls, path: Path) -> "BaseArray":
        classes = (WKWArray, ZarritaArray, ZarrArray)
        for cls in classes:
            try:
                array = cls.open(path)
                return array
            except ArrayException:  # noqa: PERF203 `try`-`except` within a loop incurs performance overhead
                pass
        raise ArrayException(f"Could not open the array at {path}.")

    @classmethod
    @abstractmethod
    def create(cls, path: Path, array_info: ArrayInfo) -> "BaseArray":
        pass

    @abstractmethod
    def read(self, bbox: NDBoundingBox) -> np.ndarray:
        pass

    @abstractmethod
    def write(self, bbox: NDBoundingBox, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def ensure_size(
        self,
        new_bbox: NDBoundingBox,
        align_with_shards: bool = True,
        warn: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def list_bounding_boxes(self) -> Iterator[NDBoundingBox]:
        "The bounding boxes are measured in voxels of the current mag."

    @abstractmethod
    def close(self) -> None:
        pass

    @staticmethod
    def get_class(data_format: DataFormat) -> Type["BaseArray"]:
        if data_format == DataFormat.WKW:
            return WKWArray
        if data_format == DataFormat.Zarr3:
            return ZarritaArray
        if data_format == DataFormat.Zarr:
            return ZarritaArray
        raise ValueError(f"Array format `{data_format}` is invalid.")


class WKWArray(BaseArray):
    data_format = DataFormat.WKW

    _cached_wkw_dataset: Optional[wkw.Dataset]

    def __init__(self, path: Path):
        super().__init__(path)
        self._cached_wkw_dataset = None

    @classmethod
    def open(cls, path: Path) -> "WKWArray":
        header_path = path / "header.wkw"

        if header_path.exists() and header_path.is_file():
            return cls(path)
        raise ArrayException(
            f"Could not open WKW array at {path}. `header.wkw` not found."
        )

    @property
    def info(self) -> ArrayInfo:
        header = self._wkw_dataset.header
        return ArrayInfo(
            data_format=self.data_format,
            num_channels=header.num_channels,
            voxel_type=header.voxel_type,
            compression_mode=header.block_type != wkw.Header.BLOCK_TYPE_RAW,
            chunk_shape=Vec3Int.full(header.block_len),
            chunks_per_shard=Vec3Int.full(
                header.file_len,
            ),
        )

    @classmethod
    def create(cls, path: Path, array_info: ArrayInfo) -> "WKWArray":
        assert array_info.data_format == cls.data_format

        assert array_info.chunk_shape.is_uniform(), f"`chunk_shape` needs to be uniform for WKW storage. Got {array_info.chunk_shape}."
        assert _is_power_of_two(
            array_info.chunk_shape.x
        ), f"`chunk_shape` needs to be a power of 2 for WKW storage. Got {array_info.chunk_shape.x}."
        assert (
            1 <= array_info.chunk_shape.x and array_info.chunk_shape.x <= 32768
        ), f"`chunk_shape` needs to be a value between 1 and 32768 for WKW storage. Got {array_info.chunk_shape.x}."

        assert array_info.chunks_per_shard.is_uniform(), f"`chunks_per_shard` needs to be uniform for WKW storage. Got {array_info.chunks_per_shard}."
        assert _is_power_of_two(
            array_info.chunks_per_shard.x
        ), f"`chunks_per_shard` needs to be a power of 2 for WKW storage. Got {array_info.chunks_per_shard.x}."
        assert (
            1 <= array_info.chunks_per_shard.x
            and array_info.chunks_per_shard.x <= 32768
        ), f"`chunks_per_shard` needs to be a value between 1 and 32768 for WKW storage. Got {array_info.chunks_per_shard.x}."

        try:
            wkw.Dataset.create(
                str(path),
                wkw.Header(
                    voxel_type=array_info.voxel_type,
                    num_channels=array_info.num_channels,
                    block_len=array_info.chunk_shape.x,
                    file_len=array_info.chunks_per_shard.x,
                    block_type=(
                        wkw.Header.BLOCK_TYPE_LZ4HC
                        if array_info.compression_mode
                        else wkw.Header.BLOCK_TYPE_RAW
                    ),
                ),
            ).close()
        except wkw.wkw.WKWException as e:
            raise ArrayException(f"Exception while creating array {path}") from e
        return WKWArray(path)

    def read(self, bbox: NDBoundingBox) -> np.ndarray:
        return self._wkw_dataset.read(Vec3Int(bbox.topleft), Vec3Int(bbox.size))

    def write(self, bbox: NDBoundingBox, data: np.ndarray) -> None:
        self._wkw_dataset.write(Vec3Int(bbox.topleft), data)

    def ensure_size(
        self,
        new_bbox: NDBoundingBox,
        align_with_shards: bool = True,
        warn: bool = False,
    ) -> None:
        pass

    def _list_files(self) -> Iterator[Path]:
        return (
            Path(relpath(filename, self._path))
            for filename in self._wkw_dataset.list_files()
        )

    def list_bounding_boxes(self) -> Iterator[NDBoundingBox]:
        def _extract_num(s: str) -> int:
            match = re.search("[0-9]+", s)
            assert match is not None
            return int(match[0])

        def _extract_file_index(file_path: Path) -> Vec3Int:
            z, y, x = [_extract_num(el) for el in file_path.parts]
            return Vec3Int(x, y, z)

        shard_shape = self.info.shard_shape
        for file_path in self._list_files():
            cube_index = _extract_file_index(file_path)
            cube_offset = cube_index * shard_shape

            yield BoundingBox(cube_offset, shard_shape)

    def close(self) -> None:
        if self._cached_wkw_dataset is not None:
            self._cached_wkw_dataset.close()
            self._cached_wkw_dataset = None

    @property
    def _wkw_dataset(self) -> wkw.Dataset:
        if self._cached_wkw_dataset is None:
            try:
                self._cached_wkw_dataset = wkw.Dataset.open(
                    str(self._path)
                )  # No need to pass the header to the wkw.Dataset
            except wkw.wkw.WKWException as e:
                raise ArrayException(
                    f"Exception while opening WKW array for {self._path}"
                ) from e
        return self._cached_wkw_dataset

    @_wkw_dataset.deleter
    def _wkw_dataset(self) -> None:
        self.close()

    def __del__(self) -> None:
        del self._cached_wkw_dataset

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_cached_wkw_dataset"]
        return d

    def __setstate__(self, d: Dict[str, Any]) -> None:
        d["_cached_wkw_dataset"] = None
        self.__dict__ = d


class ZarrArray(BaseArray):
    data_format = DataFormat.Zarr

    _cached_zarray: Optional[zarr.Array]

    def __init__(self, path: Path):
        super().__init__(path)
        self._cached_zarray = None

    @classmethod
    def open(cls, path: Path) -> "ZarrArray":
        zarray_path = path / ".zarray"

        if zarray_path.exists() and zarray_path.is_file():
            return cls(path)
        raise ArrayException(
            f"Could not open Zarr array at {path}. `.zarray` not found."
        )

    @property
    def info(self) -> ArrayInfo:
        zarray = self._zarray
        return ArrayInfo(
            data_format=self.data_format,
            num_channels=zarray.shape[0],
            voxel_type=zarray.dtype,
            compression_mode=zarray.compressor is not None,
            chunk_shape=Vec3Int(*zarray.chunks[1:4]) or Vec3Int.full(1),
            chunks_per_shard=Vec3Int.full(1),
        )

    @classmethod
    def create(cls, path: Path, array_info: ArrayInfo) -> "ZarrArray":
        assert array_info.data_format == cls.data_format
        assert array_info.chunks_per_shard == Vec3Int.full(
            1
        ), "Zarr storage doesn't support sharding yet"
        zarr.create(
            shape=(array_info.num_channels, 1, 1, 1),
            chunks=(array_info.num_channels,) + array_info.chunk_shape.to_tuple(),
            dtype=array_info.voxel_type,
            compressor=(
                numcodecs.Blosc(cname="zstd", clevel=3, shuffle=numcodecs.Blosc.SHUFFLE)
                if array_info.compression_mode
                else None
            ),
            store=_fsstore_from_path(path),
            order="F",
        )
        return ZarrArray(path)

    def read(self, bbox: NDBoundingBox) -> np.ndarray:
        shape = bbox.size
        zarray = self._zarray
        with _blosc_disable_threading():
            data = zarray[(slice(None),) + bbox.to_slices()]

        shape_with_channels = (self.info.num_channels,) + shape.to_tuple()
        if data.shape not in (shape, shape_with_channels):
            padded_data = np.zeros(shape_with_channels, dtype=data.dtype)
            padded_data[
                :,
                0 : data.shape[1],
                0 : data.shape[2],
                0 : data.shape[3],
            ] = data
            data = padded_data
        return data

    def ensure_size(
        self,
        new_bbox: NDBoundingBox,
        align_with_shards: bool = True,
        warn: bool = False,
    ) -> None:
        new_shape = VecInt(new_bbox.size, axes=new_bbox.axes)
        zarray = self._zarray

        new_shape_tuple = (zarray.shape[0],) + tuple(
            (
                max(zarray.shape[i + 1], new_shape[i])
                if len(zarray.shape) > i
                else new_shape[i]
            )
            for i in range(len(new_shape))
        )
        if new_shape_tuple != zarray.shape:
            if align_with_shards:
                shard_shape = self.info.shard_shape
                new_shape = new_shape.ceildiv(shard_shape) * shard_shape
                new_shape_tuple = (zarray.shape[0],) + new_shape.to_tuple()

            # Check on-disk for changes to shape
            current_zarray = zarr.open_array(
                store=_fsstore_from_path(self._path), mode="r"
            )
            if zarray.shape != current_zarray.shape:
                warnings.warn(
                    f"[WARNING] While resizing the Zarr array at {self._path}, a differing shape ({zarray.shape} != {current_zarray.shape}) was found in the currently persisted metadata."
                    + "This is likely happening because multiple processes changed the metadata of this array."
                )

            if warn:
                warnings.warn(
                    f"[WARNING] Resizing zarr array from `{zarray.shape}` to `{new_shape_tuple}`."
                )
            zarray.resize(new_shape_tuple)

    def write(self, bbox: NDBoundingBox, data: np.ndarray) -> None:
        """Writes a ZarrArray. If offset and bbox are given, the bbox is preferred to enable writing of n-dimensional data."""

        # If data is 3-dimensional, it is assumed that num_channels=1.
        if data.ndim == 3:
            data = data.reshape((1,) + data.shape)
        assert data.ndim == 4

        with _blosc_disable_threading():
            self.ensure_size(bbox, warn=True)
            zarray = self._zarray
            index_tuple = (slice(None),) + bbox.to_slices()

            zarray[index_tuple] = data

    def list_bounding_boxes(self) -> Iterator[NDBoundingBox]:
        zarray = self._zarray
        chunk_shape = Vec3Int(*zarray.chunks[1:4])
        for key in zarray.store.keys():
            if not key.startswith("."):
                key_parts = [int(p) for p in key.split(zarray._dimension_separator)]
                chunk_idx = Vec3Int(key_parts[1:4])
                yield BoundingBox(topleft=chunk_idx * chunk_shape, size=chunk_shape)

    def close(self) -> None:
        if self._cached_zarray is not None:
            self._cached_zarray = None

    @property
    def _zarray(self) -> zarr.Array:
        if self._cached_zarray is None:
            try:
                self._cached_zarray = zarr.open_array(
                    store=_fsstore_from_path(self._path), mode="a"
                )
            except Exception as e:
                raise ArrayException(
                    f"Exception while opening Zarr array for {self._path}"
                ) from e
        return self._cached_zarray

    @_zarray.deleter
    def _zarray(self) -> None:
        self.close()

    def __del__(self) -> None:
        del self._cached_zarray

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_cached_zarray"]
        return d

    def __setstate__(self, d: Dict[str, Any]) -> None:
        d["_cached_zarray"] = None
        self.__dict__ = d


class ZarritaArray(BaseArray):
    data_format = DataFormat.Zarr3

    _cached_zarray: Optional[Union["zarrita.Array", "zarrita.ArrayV2"]]

    def __init__(self, path: Path):
        super().__init__(path)
        self._cached_zarray = None

    @classmethod
    def open(cls, path: Path) -> "ZarritaArray":
        from zarrita import Array

        try:
            Array.open_auto(store=path)  # check that everything exists
            return cls(path)
        except Exception as exc:
            raise ArrayException(f"Could not open Zarr array at {path}.") from exc

    @staticmethod
    def _has_compression_codecs(codecs: List["zarrita.codecs.Codec"]) -> bool:
        from zarrita.codecs import BloscCodec, GzipCodec, ZstdCodec

        return any(
            isinstance(c, BloscCodec)
            or isinstance(c, GzipCodec)
            or isinstance(c, ZstdCodec)
            for c in codecs
        )

    @property
    def info(self) -> ArrayInfo:
        from zarrita import Array
        from zarrita.sharding import ShardingCodec

        zarray = self._zarray
        if (names := getattr(zarray.metadata, "dimension_names", None)) is None:
            dimension_names = ("c", "x", "y", "z")
        else:
            dimension_names = names
        if isinstance(zarray, Array):
            if len(zarray.codec_pipeline.codecs) == 1 and isinstance(
                zarray.codec_pipeline.codecs[0], ShardingCodec
            ):
                sharding_codec = zarray.codec_pipeline.codecs[0]
                shard_shape = zarray.metadata.chunk_grid.configuration.chunk_shape
                chunk_shape = sharding_codec.configuration.chunk_shape
                return ArrayInfo(
                    data_format=DataFormat.Zarr3,
                    num_channels=zarray.metadata.shape[0],
                    voxel_type=zarray.metadata.dtype,
                    compression_mode=self._has_compression_codecs(
                        sharding_codec.codec_pipeline.codecs
                    ),
                    chunk_shape=Vec3Int(chunk_shape[1:4]),
                    chunks_per_shard=Vec3Int(
                        Vec3Int(shard_shape[1:4]) // Vec3Int(chunk_shape[1:4])
                    ),
                    dimension_names=dimension_names,
                )
            return ArrayInfo(
                data_format=DataFormat.Zarr3,
                num_channels=zarray.metadata.shape[0],
                voxel_type=zarray.metadata.dtype,
                compression_mode=self._has_compression_codecs(
                    zarray.codec_pipeline.codecs
                ),
                chunk_shape=Vec3Int(
                    zarray.metadata.chunk_grid.configuration.chunk_shape[1:4]
                )
                or Vec3Int.full(1),
                chunks_per_shard=Vec3Int.full(1),
                dimension_names=dimension_names,
            )
        else:
            return ArrayInfo(
                data_format=DataFormat.Zarr,
                num_channels=zarray.metadata.shape[0],
                voxel_type=zarray.metadata.dtype,
                compression_mode=zarray.metadata.compressor is not None,
                chunk_shape=Vec3Int(*zarray.metadata.chunks[1:4]) or Vec3Int.full(1),
                chunks_per_shard=Vec3Int.full(1),
                dimension_names=dimension_names,
            )

    @classmethod
    def create(cls, path: Path, array_info: ArrayInfo) -> "ZarritaArray":
        import zarrita.codecs
        from zarrita import Array, ArrayV2

        assert array_info.data_format in (DataFormat.Zarr, DataFormat.Zarr3)
        if array_info.data_format == DataFormat.Zarr3:
            chunk_shape = (array_info.num_channels,) + tuple(
                getattr(array_info.chunk_shape, axis, 1)
                for axis in array_info.dimension_names[1:]
            )
            shard_shape = (array_info.num_channels,) + tuple(
                getattr(array_info.shard_shape, axis, 1)
                for axis in array_info.dimension_names[1:]
            )
            Array.create(
                store=path,
                shape=array_info.shape,
                chunk_shape=shard_shape,
                chunk_key_encoding=("default", "/"),
                dtype=array_info.voxel_type,
                dimension_names=array_info.dimension_names,
                codecs=[
                    zarrita.codecs.sharding_codec(
                        chunk_shape=chunk_shape,
                        codecs=(
                            [
                                zarrita.codecs.transpose_codec(array_info.axis_order),
                                zarrita.codecs.bytes_codec(),
                                zarrita.codecs.blosc_codec(
                                    typesize=array_info.voxel_type.itemsize
                                ),
                            ]
                            if array_info.compression_mode
                            else [
                                zarrita.codecs.transpose_codec(array_info.axis_order),
                                zarrita.codecs.bytes_codec(),
                            ]
                        ),
                    )
                ],
            )
        else:
            ArrayV2.create(
                store=path,
                shape=(array_info.shape),
                chunks=(array_info.num_channels,) + array_info.chunk_shape.to_tuple(),
                dtype=array_info.voxel_type,
                compressor=(
                    {"id": "blosc", "cname": "zstd", "clevel": 5}
                    if array_info.compression_mode
                    else None
                ),
                order="F",
                dimension_separator=".",
            )
        return ZarritaArray(path)

    def read(self, bbox: NDBoundingBox) -> np.ndarray:
        shape = bbox.size.to_tuple()
        zarray = self._zarray
        slice_tuple = (slice(None),) + bbox.to_slices()
        with _blosc_disable_threading():
            data = zarray[slice_tuple]

        shape_with_channels = (self.info.num_channels,) + shape
        if data.shape != shape_with_channels:
            data_slice_tuple = tuple(slice(0, size) for size in data.shape)
            padded_data = np.zeros(shape_with_channels, dtype=zarray.metadata.dtype)
            padded_data[data_slice_tuple] = data
            data = padded_data
        return data

    def ensure_size(
        self,
        new_bbox: NDBoundingBox,
        align_with_shards: bool = True,
        warn: bool = False,
    ) -> None:
        zarray = self._zarray

        new_bbox = new_bbox.with_bottomright(
            (
                max(zarray.metadata.shape[i + 1], new_bbox.bottomright[i])
                for i in range(len(new_bbox))
            )
        )
        new_shape_tuple = (zarray.metadata.shape[0],) + tuple(new_bbox.bottomright)
        if new_shape_tuple != zarray.metadata.shape:
            if align_with_shards:
                shard_shape = self.info.shard_shape
                new_aligned_bbox = new_bbox.with_bottomright_xyz(
                    new_bbox.bottomright_xyz.ceildiv(shard_shape) * shard_shape
                )
                new_shape_tuple = (
                    zarray.metadata.shape[0],
                ) + new_aligned_bbox.bottomright.to_tuple()

            # Check on-disk for changes to shape
            current_zarray = zarray.open(self._path)
            if zarray.metadata.shape != current_zarray.metadata.shape:
                warnings.warn(
                    f"[WARNING] While resizing the Zarr array at {self._path}, a differing shape ({zarray.metadata.shape} != {current_zarray.metadata.shape}) was found in the currently persisted metadata."
                    + "This is likely happening because multiple processes changed the metadata of this array."
                )

            if warn:
                warnings.warn(
                    f"[WARNING] Resizing zarr array from `{zarray.metadata.shape}` to `{new_shape_tuple}`."
                )
            self._cached_zarray = zarray.resize(new_shape_tuple)

    def write(self, bbox: NDBoundingBox, data: np.ndarray) -> None:
        if data.ndim == len(bbox):
            # the bbox does not include the channels, if data and bbox have the same size there is only 1 channel
            data = data.reshape((1,) + data.shape)

        assert data.ndim == len(bbox) + 1

        with _blosc_disable_threading():
            self.ensure_size(bbox, warn=True)
            zarray = self._zarray
            index_tuple = (slice(None),) + bbox.to_slices()

            zarray[index_tuple] = data

    def list_bounding_boxes(self) -> Iterator[NDBoundingBox]:
        raise NotImplementedError

    def close(self) -> None:
        if self._cached_zarray is not None:
            self._cached_zarray = None

    @property
    def _zarray(self) -> Union["zarrita.Array", "zarrita.ArrayV2"]:
        from zarrita import Array, runtime_configuration

        if self._cached_zarray is None:
            try:
                zarray = Array.open_auto(
                    store=self._path,
                    runtime_configuration=runtime_configuration("F"),
                )
                self._cached_zarray = zarray
            except Exception as e:
                raise ArrayException(
                    f"Exception while opening Zarrita array for {self._path}"
                ) from e
        return self._cached_zarray

    @_zarray.deleter
    def _zarray(self) -> None:
        self.close()

    def __del__(self) -> None:
        del self._cached_zarray

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_cached_zarray"]
        return d

    def __setstate__(self, d: Dict[str, Any]) -> None:
        d["_cached_zarray"] = None
        self.__dict__ = d



from typing import Dict, Optional, Tuple

from webknossos.client.api_client.models import (
    ApiDatasetUploadInformation,
    ApiReserveDatasetUploadInformation,
)

from ._abstract_api_client import LONG_TIMEOUT_SECONDS, AbstractApiClient, Query


class DatastoreApiClient(AbstractApiClient):
    # Client to use the HTTP API of WEBKNOSSOS datastore servers.
    # When adding a method here, use the utility methods from AbstractApiClient
    # and add more as needed.
    # Methods here are prefixed with the domain, e.g. dataset_finish_upload (not finish_dataset_upload)

    def __init__(
        self,
        datastore_base_url: str,
        timeout_seconds: float,
        headers: Optional[Dict[str, str]] = None,
    ):
        super().__init__(timeout_seconds, headers)
        self.datastore_base_url = datastore_base_url

    @property
    def url_prefix(self) -> str:
        return f"{self.datastore_base_url}/data"

    def dataset_finish_upload(
        self,
        upload_information: ApiDatasetUploadInformation,
        token: Optional[str],
        retry_count: int,
    ) -> None:
        route = "/datasets/finishUpload"
        return self._post_json(
            route,
            upload_information,
            query={"token": token},
            retry_count=retry_count,
            timeout_seconds=LONG_TIMEOUT_SECONDS,
        )

    def dataset_reserve_upload(
        self,
        reserve_upload_information: ApiReserveDatasetUploadInformation,
        token: Optional[str],
        retry_count: int,
    ) -> None:
        route = "/datasets/reserveUpload"
        return self._post_json(
            route,
            reserve_upload_information,
            query={"token": token},
            retry_count=retry_count,
        )

    def dataset_get_raw_data(
        self,
        organization_name: str,
        dataset_name: str,
        data_layer_name: str,
        mag: str,
        token: Optional[str],
        x: int,
        y: int,
        z: int,
        width: int,
        height: int,
        depth: int,
    ) -> Tuple[bytes, str]:
        route = f"/datasets/{organization_name}/{dataset_name}/layers/{data_layer_name}/data"
        query: Query = {
            "mag": mag,
            "x": x,
            "y": y,
            "z": z,
            "width": width,
            "height": height,
            "depth": depth,
            "token": token,
        }
        response = self._get(route, query)
        return response.content, response.headers.get("MISSING-BUCKETS")



from collections.abc import MutableMapping
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Tuple, Union

import networkx as nx
import numpy as np

from ..geometry import Vec3Int, Vec3IntLike
from ..utils import warn_deprecated
from .node import Node

if TYPE_CHECKING:
    from .group import Group
    from .skeleton import Skeleton

Vector3 = Tuple[float, float, float]
Vector4 = Tuple[float, float, float, float]


def _get_id(node_or_id: Union[Node, int]) -> int:
    if isinstance(node_or_id, Node):
        return node_or_id.id
    else:
        return node_or_id


class _NodeDict(MutableMapping):
    """Dict-like container for the `Tree` class below to handle both nodes and ids as keys.
    Only when setting the value of a node for the first time the actual node must be passed.
    This allows to keep a reference from all ids to the nodes, which can be looked up
    using `get_node()`. Iterating over the keys yields the `Node` objects."""

    def __init__(self) -> None:
        self._id_to_attrs: Dict[int, Any] = {}
        self._id_to_node: Dict[int, Node] = {}

    def __getitem__(self, key: Union[Node, int]) -> Any:
        return self._id_to_attrs[_get_id(key)]

    def __setitem__(self, key: Union[Node, int], value: Dict) -> None:
        key_id = _get_id(key)
        if key_id not in self._id_to_node:
            if isinstance(key, Node):
                self._id_to_node[key_id] = key
            else:
                raise ValueError(
                    f"Tried to add node {key}, which does not exist yet, to a tree."
                    + "For insertion, the node must be an instance of the Node class."
                )
        self._id_to_attrs[key_id] = value

    def __delitem__(self, key: Union[Node, int]) -> None:
        del self._id_to_node[_get_id(key)]
        del self._id_to_attrs[_get_id(key)]

    def __iter__(self) -> Iterator[Node]:
        return iter(self._id_to_node.values())

    def __len__(self) -> int:
        return len(self._id_to_attrs)

    def get_node(self, id_: int) -> Node:
        return self._id_to_node[id_]


class _AdjDict(MutableMapping):
    """Dict-like container for the `Tree` class below to handle both nodes and ids as keys.
    Needs a reference to the _node attribute (of class `_NodeDict`) of the graph object
    to get a reference from ids to nodes. Iterating over the keys yields the `Node` objects.

    See Tree.__init__ for more details"""

    def __init__(self, *, node_dict: _NodeDict) -> None:
        self._id_to_attrs: Dict[int, Any] = {}
        self._node_dict = node_dict

    def __getitem__(self, key: Union[Node, int]) -> Any:
        return self._id_to_attrs[_get_id(key)]

    def __setitem__(self, key: Union[Node, int], value: Dict) -> None:
        self._id_to_attrs[_get_id(key)] = value

    def __delitem__(self, key: Union[Node, int]) -> None:
        del self._id_to_attrs[_get_id(key)]

    def __iter__(self) -> Iterator[Node]:
        return (self._node_dict.get_node(i) for i in self._id_to_attrs)

    def __len__(self) -> int:
        return len(self._id_to_attrs)


class Tree(nx.Graph):
    """
    Contains a collection of nodes and edges. Despite the name, trees may contain cycles.
    This class inherits from [`networkx.Graph`](https://networkx.org/documentation/stable/reference/classes/graph.html).
    For further methods, please [check the networkx documentation](https://networkx.org/documentation/stable/reference/classes/graph.html#methods).

    See Tree.__init__ for more details.

    A small usage example:

    ```python
    tree = skeleton.add_tree("a tree")
    node_1 = tree.add_node(position=(0, 0, 0), comment="node 1")
    node_2 = tree.add_node(position=(100, 100, 100), comment="node 2")

    tree.add_edge(node_1, node_2)
    ```
    """

    def __init__(
        self,
        name: str,
        group: "Group",
        skeleton: "Skeleton",
        color: Optional[Vector4] = None,
        enforced_id: Optional[int] = None,  # noqa: ARG002 Unused method argument: `enforced_id`
    ) -> None:
        """
        To create a tree, it is recommended to use `Skeleton.add_tree` or
        `Group.add_tree`. That way, the newly created tree is automatically
        attached as a child to the object the method was called on.
        """

        super().__init__()
        # Note that id is set up in __new__
        self.name = name
        self.group = group
        self.color = color

        # only used internally
        self._skeleton = skeleton

    def __new__(
        cls,
        name: str,  # noqa: ARG003 Unused class method argument: `name`
        group: "Group",  # noqa: ARG003 Unused class method argument: `group`
        skeleton: "Skeleton",
        color: Optional[Vector4] = None,  # noqa: ARG003 Unused class method argument: `color`
        enforced_id: Optional[int] = None,
    ) -> "Tree":
        self = super().__new__(cls)

        # self._id is a read-only member, exposed via properties.
        # It is set in __new__ instead of __init__ so that pickling/unpickling
        # works without problems. As long as the deserialization of a tree instance
        # is not finished, the object is only half-initialized. Since self._id
        # is needed by __hash__, an error would be raised otherwise.
        # Also see:
        # https://stackoverflow.com/questions/46283738/attributeerror-when-using-python-deepcopy
        if enforced_id is not None:
            self._id = enforced_id
        else:
            self._id = skeleton._element_id_generator.__next__()

        return self

    def __getnewargs__(self) -> Tuple:
        # pickle.dump will pickle instances of Tree so that the following
        # tuple is passed as arguments to __new__.
        return (self.name, self.group, self._skeleton, self.color, self._id)

    # node_dict_factory, adjlist_outer_dict_factory and adjlist_inner_dict_factory are used by networkx
    # from which we subclass.
    # To be able to reference nodes by id after adding them for the first time, we use custom dict-like classes
    # for the networkx-graph structures, that have nodes as keys:
    #     * `self._node`: _NodeDict
    #        holding the attributes of nodes, keeping references from ids to nodes
    #     * `self._adj`: _AdjDict on the first two levels
    #       holding edge attributes on the last level, using self._node to convert ids to nodes
    # It's important to set the attributes before the parent's init so that they shadow the class-attributes.
    # For further details, see the *Subclasses* section here: https://networkx.org/documentation/stable/reference/classes/graph.html

    node_dict_factory = _NodeDict

    def adjlist_outer_dict_factory(self) -> _AdjDict:
        # self._node will already be available when this method is called, because networkx.Graph.__init__
        # sets up the nodes first and then the edges (i.e., adjacency list).
        return _AdjDict(node_dict=self._node)

    def adjlist_inner_dict_factory(self) -> _AdjDict:
        return _AdjDict(node_dict=self._node)

    def __to_tuple_for_comparison(self) -> Tuple:
        return (
            self.name,
            self.id,
            self.color,
            sorted(self.nodes),
            sorted(self.edges),
        )

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Tree) and (
            self.__to_tuple_for_comparison() == o.__to_tuple_for_comparison()
        )

    @property
    def id(self) -> int:
        """Read-only property."""
        return self._id

    def get_node_positions(self) -> np.ndarray:
        """Returns an numpy array with the positions of all nodes of this tree."""
        return np.array([node.position for node in self.nodes])

    def get_node_by_id(self, node_id: int) -> Node:
        """Returns the node in this tree with the requested id."""
        return self._node.get_node(node_id)

    def add_node(
        self,
        position: Vec3IntLike,
        comment: Optional[str] = None,
        radius: Optional[float] = None,
        rotation: Optional[Vector3] = None,
        inVp: Optional[int] = None,
        inMag: Optional[int] = None,
        bitDepth: Optional[int] = None,
        interpolation: Optional[bool] = None,
        time: Optional[int] = None,
        is_branchpoint: bool = False,
        branchpoint_time: Optional[int] = None,
        _enforced_id: Optional[int] = None,
    ) -> Node:
        """
        Adds a node to the tree. Apart from the mandatory `position` parameter,
        there are several optional parameters which can be used to encode
        additional information. For example, the comment will be shown by the
        WEBKNOSSOS UI.
        """
        node = Node(
            position=Vec3Int(position),
            comment=comment,
            radius=radius,
            rotation=rotation,
            inVp=inVp,
            inMag=inMag,
            bitDepth=bitDepth,
            interpolation=interpolation,
            time=time,
            is_branchpoint=is_branchpoint,
            branchpoint_time=branchpoint_time,
            enforced_id=_enforced_id,
            skeleton=self._skeleton,
        )
        super().add_node(node)
        return node

    def get_max_node_id(self) -> int:
        """Returns the highest node id."""
        return max((node.id for node in self.nodes), default=0)

    def __hash__(self) -> int:
        return self._id


class Graph(Tree):
    """Deprecated, please use `Tree` instead."""

    def __init__(
        self,
        name: str,
        group: "Group",
        skeleton: "Skeleton",
        color: Optional[Vector4] = None,
        enforced_id: Optional[int] = None,
    ) -> None:
        warn_deprecated("Graph", "Tree")
        super().__init__(
            name=name,
            group=group,
            skeleton=skeleton,
            color=color,
            enforced_id=enforced_id,
        )
